@@ -49,7 +49,7 @@ defmodule Exception do
   @doc """
   Receives the exception struct and must return its message.
 
-  Most commonly exceptions have a message field which by default is accessed
+  Many exceptions have a message field which by default is accessed
   by this function. However, if an exception does not have a message field,
   this function must be explicitly implemented.
   """
@@ -346,7 +346,7 @@ defmodule Exception do
   defp rewrite_arg(arg) do
     Macro.prewalk(arg, fn
       {:%{}, meta, [__struct__: Range, first: first, last: last, step: step]} ->
-        {:"..//", meta, [first, last, step]}
+        {:..//, meta, [first, last, step]}
 
       other ->
         other
@@ -407,8 +407,8 @@ defmodule Exception do
 
   defp erl_to_ex(mod, fun, args, meta) do
     case :elixir_rewrite.erl_to_ex(mod, fun, args) do
-      {Kernel, fun, args} -> {fun, meta, args}
-      {mod, fun, args} -> {{:., [], [mod, fun]}, meta, args}
+      {Kernel, fun, args, _} -> {fun, meta, args}
+      {mod, fun, args, _} -> {{:., [], [mod, fun]}, meta, args}
     end
   end
 
@@ -1581,13 +1581,15 @@ defmodule UndefinedFunctionError do
 
   """
 
+  @function_threshold 0.77
+  @max_suggestions 5
   defexception [:module, :function, :arity, :reason, :message]
 
   @impl true
   def message(%{message: nil} = exception) do
     %{reason: reason, module: module, function: function, arity: arity} = exception
-    {message, _hint_type} = message(reason, module, function, arity)
-    message
+    {message, hint_type} = message(reason, module, function, arity)
+    message <> if(hint_type == :suggest_module, do: hint_for_missing_alias(module), else: "")
   end
 
   def message(%{message: message}) do
@@ -1648,12 +1650,25 @@ defmodule UndefinedFunctionError do
   end
 
   defp hint(module, function, arity, :suggest_function) do
-    behaviour_hint(module, function, arity) <>
-      hint_for_loaded_module(module, function, arity, nil)
+    hint_for_behaviour(module, function, arity) <>
+      hint_for_loaded_module(module, function, arity)
   end
 
-  @max_suggestions 5
   defp hint(module, function, arity, :suggest_module) do
+    hint_for_missing_module(module, function, arity)
+  end
+
+  defp hint_for_missing_alias(module) do
+    with "Elixir." <> rest <- Atom.to_string(module),
+         false <- rest =~ "." do
+      ". Make sure the module name is correct and has been specified in full (or that an alias has been defined)"
+    else
+      _ -> ""
+    end
+  end
+
+  @doc false
+  def hint_for_missing_module(module, function, arity) do
     downcased_module = downcase_module_name(module)
     stripped_module = module |> Atom.to_string() |> String.replace_leading("Elixir.", "")
 
@@ -1674,9 +1689,9 @@ defmodule UndefinedFunctionError do
           ["\n      * ", Exception.format_mfa(module, function, arity)]
         end)
 
-      IO.iodata_to_binary([". Did you mean:\n", suggestions, "\n"])
+      ". Did you mean:\n#{suggestions}\n"
     else
-      ""
+      hint_for_missing_alias(module)
     end
   end
 
@@ -1693,7 +1708,7 @@ defmodule UndefinedFunctionError do
   end
 
   @doc false
-  def hint_for_loaded_module(module, function, arity, exports) do
+  def hint_for_loaded_module(module, function, arity) do
     cond do
       macro_exported?(module, function, arity) ->
         ". However, there is a macro with the same name and arity. " <>
@@ -1703,7 +1718,7 @@ defmodule UndefinedFunctionError do
         ", #{message}"
 
       true ->
-        IO.iodata_to_binary(did_you_mean(module, function, exports))
+        IO.iodata_to_binary(did_you_mean(module, function))
     end
   end
 
@@ -1714,11 +1729,39 @@ defmodule UndefinedFunctionError do
     end
   end
 
-  @function_threshold 0.77
-  @max_suggestions 5
+  defp hint_for_behaviour(module, function, arity) do
+    case behaviours_for(module) do
+      [] ->
+        ""
 
-  defp did_you_mean(module, function, exports) do
-    exports = exports || exports_for(module)
+      behaviours ->
+        case Enum.find(behaviours, &expects_callback?(&1, function, arity)) do
+          nil -> ""
+          behaviour -> ", but the behaviour #{inspect(behaviour)} expects it to be present"
+        end
+    end
+  rescue
+    # In case the module was removed while we are computing this
+    UndefinedFunctionError -> ""
+  end
+
+  defp behaviours_for(module) do
+    :attributes
+    |> module.module_info()
+    |> Keyword.get(:behaviour, [])
+  end
+
+  defp expects_callback?(behaviour, function, arity) do
+    callbacks =
+      behaviour.behaviour_info(:callbacks) -- behaviour.behaviour_info(:optional_callbacks)
+
+    Enum.member?(callbacks, {function, arity})
+  end
+
+  ## Shared helpers across hints
+
+  defp did_you_mean(module, function) do
+    exports = exports_for(module)
 
     result =
       case Keyword.take(exports, [function]) do
@@ -1748,35 +1791,6 @@ defmodule UndefinedFunctionError do
     ["      * ", Macro.inspect_atom(:remote_call, fun), ?/, Integer.to_string(arity), ?\n]
   end
 
-  defp behaviour_hint(module, function, arity) do
-    case behaviours_for(module) do
-      [] ->
-        ""
-
-      behaviours ->
-        case Enum.find(behaviours, &expects_callback?(&1, function, arity)) do
-          nil -> ""
-          behaviour -> ", but the behaviour #{inspect(behaviour)} expects it to be present"
-        end
-    end
-  rescue
-    # In case the module was removed while we are computing this
-    UndefinedFunctionError -> ""
-  end
-
-  defp behaviours_for(module) do
-    :attributes
-    |> module.module_info()
-    |> Keyword.get(:behaviour, [])
-  end
-
-  defp expects_callback?(behaviour, function, arity) do
-    callbacks =
-      behaviour.behaviour_info(:callbacks) -- behaviour.behaviour_info(:optional_callbacks)
-
-    Enum.member?(callbacks, {function, arity})
-  end
-
   defp exports_for(module) do
     if function_exported?(module, :__info__, 1) do
       module.__info__(:macros) ++ module.__info__(:functions)
@@ -1785,8 +1799,7 @@ defmodule UndefinedFunctionError do
     end
   rescue
     # In case the module was removed while we are computing this
-    UndefinedFunctionError ->
-      []
+    UndefinedFunctionError -> []
   end
 
   defp deprecated_functions_for(module) do
@@ -1797,8 +1810,7 @@ defmodule UndefinedFunctionError do
     end
   rescue
     # In case the module was removed while we are computing this
-    UndefinedFunctionError ->
-      []
+    UndefinedFunctionError -> []
   end
 end
 
@@ -1985,8 +1997,23 @@ defmodule Protocol.UndefinedError do
 
   @impl true
   def message(%{protocol: protocol, value: value, description: description}) do
-    "protocol #{inspect(protocol)} not implemented for #{inspect(value)} of type " <>
-      value_type(value) <> maybe_description(description) <> maybe_available(protocol)
+    inspected =
+      value
+      |> inspect(pretty: true)
+      # Indent only lines with contents on them
+      |> String.replace(~r/^(?=.+)/m, "    ")
+
+    "protocol #{inspect(protocol)} not implemented for type " <>
+      value_type(value) <>
+      maybe_description(description) <>
+      maybe_available(protocol) <>
+      """
+
+
+      Got value:
+
+      #{inspected}
+      """
   end
 
   defp value_type(%{__struct__: struct}), do: "#{inspect(struct)} (a struct)"
@@ -2131,6 +2158,56 @@ defmodule UnicodeConversionError do
   end
 end
 
+defmodule MissingApplicationsError do
+  @moduledoc """
+  An exception that is raised when an application depends on one or more
+  missing applications.
+
+  This exception is used by Mix and other tools. It can also be used by library authors
+  when their library only requires an external application (like a dependency) for a subset
+  of features.
+
+  The fields of this exception are public. See `t:t/0`.
+
+  *Available since v1.18.0.*
+
+  ## Examples
+
+      unless Application.spec(:plug, :vsn) do
+        raise MissingApplicationsError,
+          description: "application :plug is required for testing Plug-related functionality",
+          apps: [{:plug, "~> 1.0"}]
+      end
+
+  """
+
+  @moduledoc since: "1.18.0"
+
+  @type t() :: %__MODULE__{
+          apps: [{Application.app(), Version.requirement()}, ...],
+          description: String.t()
+        }
+
+  defexception apps: [], description: "missing applications found"
+
+  @impl true
+  def message(%__MODULE__{apps: apps, description: description}) do
+    # We explicitly format these as tuples so that they're easier to copy-paste
+    # into dependencies.
+    formatted_apps =
+      Enum.map(apps, fn {app_name, requirement} ->
+        ~s(\n  {#{inspect(app_name)}, "#{requirement}"})
+      end)
+
+    """
+    #{description}
+
+    To address this, include these applications as your dependencies:
+    #{formatted_apps}\
+    """
+  end
+end
+
 defmodule Enum.OutOfBoundsError do
   @moduledoc """
   An exception that is raised when a function expects an enumerable to have
@@ -2453,7 +2530,7 @@ defmodule ErlangError do
           {:ok, reason, IO.iodata_to_binary([":\n\n" | Enum.map(args_errors, &arg_error/1)])}
 
         general = extra[:general] ->
-          {:ok, reason, ": " <> general}
+          {:ok, reason, ": " <> IO.chardata_to_string(general)}
 
         true ->
           :error
