@@ -39,12 +39,14 @@ debug_info(_, _, _, _) ->
 %% Builds Erlang AST annotation.
 
 get_ann(Opts) when is_list(Opts) ->
-  get_ann(Opts, false, 0).
+  get_ann(Opts, false, 0, undefined).
 
-get_ann([{generated, true} | T], _, Line) -> get_ann(T, true, Line);
-get_ann([{line, Line} | T], Gen, _) when is_integer(Line) -> get_ann(T, Gen, Line);
-get_ann([_ | T], Gen, Line) -> get_ann(T, Gen, Line);
-get_ann([], Gen, Line) -> erl_anno:set_generated(Gen, erl_anno:new(Line)).
+get_ann([{generated, true} | T], _, Line, Column) -> get_ann(T, true, Line, Column);
+get_ann([{line, Line} | T], Gen, _, Column) when is_integer(Line) -> get_ann(T, Gen, Line, Column);
+get_ann([{column, Column} | T], Gen, Line, _) when is_integer(Column) -> get_ann(T, Gen, Line, Column);
+get_ann([_ | T], Gen, Line, Column) -> get_ann(T, Gen, Line, Column);
+get_ann([], Gen, Line, undefined) -> erl_anno:set_generated(Gen, erl_anno:new(Line));
+get_ann([], Gen, Line, Column) -> erl_anno:set_generated(Gen, erl_anno:new({Line, Column})).
 
 %% Converts an Elixir definition to an anonymous function.
 
@@ -90,14 +92,35 @@ elixir_to_erl(Tree, Ann) when is_binary(Tree) ->
   %% considers a string in a binary to be encoded in latin1, so the bytes
   %% are not changed in any fashion.
   {bin, Ann, [{bin_element, Ann, {string, Ann, binary_to_list(Tree)}, default, default}]};
-elixir_to_erl(Pid, Ann) when is_pid(Pid) ->
-  ?remote(Ann, erlang, binary_to_term, [elixir_to_erl(term_to_binary(Pid), Ann)]);
-elixir_to_erl(_Other, _Ann) ->
-  error(badarg).
+elixir_to_erl(Tree, Ann) when is_bitstring(Tree) ->
+  Segments = [elixir_to_erl_bitstring_segment(X, Ann) || X <- bitstring_to_list(Tree)],
+  {bin, Ann, Segments};
+elixir_to_erl(Tree, Ann) when is_function(Tree) ->
+  case (erlang:fun_info(Tree, type) == {type, external}) andalso
+       (erlang:fun_info(Tree, env) == {env, []}) of
+    true ->
+      {module, Module} = erlang:fun_info(Tree, module),
+      {name, Name} = erlang:fun_info(Tree, name),
+      {arity, Arity} = erlang:fun_info(Tree, arity),
+      {'fun', Ann, {function, {atom, Ann, Module}, {atom, Ann, Name}, {integer, Ann, Arity}}};
+    false ->
+      error(badarg, [Tree, Ann])
+  end;
+elixir_to_erl(Tree, Ann) when is_pid(Tree); is_port(Tree); is_reference(Tree) ->
+  ?remote(Ann, erlang, binary_to_term, [elixir_to_erl(term_to_binary(Tree), Ann)]);
+elixir_to_erl(Tree, Ann) ->
+  error(badarg, [Tree, Ann]).
 
 elixir_to_erl_cons([H | T], Ann) -> {cons, Ann, elixir_to_erl(H, Ann), elixir_to_erl_cons(T, Ann)};
 elixir_to_erl_cons(T, Ann) -> elixir_to_erl(T, Ann).
 
+elixir_to_erl_bitstring_segment(Int, Ann) when is_integer(Int) ->
+  {bin_element, Ann, {integer, Ann, Int}, default, [integer]};
+elixir_to_erl_bitstring_segment(Rest, Ann) when is_bitstring(Rest) ->
+  Size = bit_size(Rest),
+  <<Int:Size>> = Rest,
+  {bin_element, Ann, {integer, Ann, Int}, {integer, Ann, Size}, [integer]}.
+
 %% Returns a scope for translation.
 
 scope(_Meta, ExpandCaptures) ->
@@ -111,7 +134,13 @@ consolidate(Map, TypeSpecs, Chunks) ->
 
 %% Dynamic compilation hook, used in regular compiler
 
-compile(#{module := Module, line := Line} = Map) ->
+compile(#{module := Module, anno := Anno} = BaseMap) ->
+  Map =
+    case elixir_erl_compiler:env_compiler_options() of
+      [] -> BaseMap;
+      EnvOptions -> BaseMap#{compile_opts := ?key(BaseMap, compile_opts) ++ EnvOptions}
+    end,
+
   {Set, Bag} = elixir_module:data_tables(Module),
 
   TranslatedTypespecs =
@@ -124,17 +153,34 @@ compile(#{module := Module, line := Line} = Map) ->
   {Prefix, Forms, Def, Defmacro, Macros} = dynamic_form(Map),
   {Types, Callbacks, TypeSpecs} = typespecs_form(Map, TranslatedTypespecs, Macros),
 
-  DocsChunk = docs_chunk(Set, Module, Line, Def, Defmacro, Types, Callbacks),
-  CheckerChunk = checker_chunk(Map),
+  ChunkOpts =
+    case lists:member(deterministic, ?key(Map, compile_opts)) of
+      true -> [deterministic];
+      false -> []
+    end,
+
+  DocsChunk = docs_chunk(Map, Set, Module, Anno, Def, Defmacro, Types, Callbacks, ChunkOpts),
+  CheckerChunk = checker_chunk(Map, Def, ChunkOpts),
   load_form(Map, Prefix, Forms, TypeSpecs, DocsChunk ++ CheckerChunk).
 
-dynamic_form(#{module := Module, line := Line, relative_file := RelativeFile,
+dynamic_form(#{module := Module, relative_file := RelativeFile,
                attributes := Attributes, definitions := Definitions, unreachable := Unreachable,
                deprecated := Deprecated, compile_opts := Opts} = Map) ->
+  %% TODO: Match on anno directly in Elixir v1.22+
+  Line = case Map of
+    #{anno := AnnoValue} -> erl_anno:line(AnnoValue);
+    #{line := LineValue} -> LineValue
+  end,
+
   {Def, Defmacro, Macros, Exports, Functions} =
     split_definition(Definitions, Unreachable, Line, [], [], [], [], {[], []}),
 
-  FilteredOpts = lists:filter(fun({no_warn_undefined, _}) -> false; (_) -> true end, Opts),
+  FilteredOpts = lists:filter(fun(
+    {no_warn_undefined, _}) -> false;
+    (debug_info) -> false;
+    (_) -> true
+  end, Opts),
+
   Location = {elixir_utils:characters_to_list(RelativeFile), Line},
 
   Prefix = [{attribute, Line, file, Location},
@@ -157,13 +203,14 @@ split_definition([{Tuple, Kind, Meta, Clauses} | T], Unreachable, Line,
     true ->
       split_definition(T, Unreachable, Line, Def, Defmacro, Macros, Exports, Functions)
   end;
+
 split_definition([], _Unreachable, _Line, Def, Defmacro, Macros, Exports, {Head, Tail}) ->
-  {Def, Defmacro, Macros, Exports, Head ++ Tail}.
+  {lists:sort(Def), lists:sort(Defmacro), Macros, Exports, Head ++ Tail}.
 
 split_definition(Tuple, def, Meta, Clauses, T, Unreachable, Line,
                  Def, Defmacro, Macros, Exports, Functions) ->
   {_, _, N, A, _} = Entry = translate_definition(def, Line, Meta, Tuple, Clauses),
-  split_definition(T, Unreachable, Line, [Tuple | Def], Defmacro, Macros, [{N, A} | Exports],
+  split_definition(T, Unreachable, Line, [{Tuple, Meta} | Def], Defmacro, Macros, [{N, A} | Exports],
                    add_definition(Meta, Entry, Functions));
 
 split_definition(Tuple, defp, Meta, Clauses, T, Unreachable, Line,
@@ -175,7 +222,7 @@ split_definition(Tuple, defp, Meta, Clauses, T, Unreachable, Line,
 split_definition(Tuple, defmacro, Meta, Clauses, T, Unreachable, Line,
                  Def, Defmacro, Macros, Exports, Functions) ->
   {_, _, N, A, _} = Entry = translate_definition(defmacro, Line, Meta, Tuple, Clauses),
-  split_definition(T, Unreachable, Line, Def, [Tuple | Defmacro], [Tuple | Macros], [{N, A} | Exports],
+  split_definition(T, Unreachable, Line, Def, [{Tuple, Meta} | Defmacro], [Tuple | Macros], [{N, A} | Exports],
                    add_definition(Meta, Entry, Functions));
 
 split_definition(Tuple, defmacrop, Meta, Clauses, T, Unreachable, Line,
@@ -247,13 +294,14 @@ is_macro(_)         -> false.
 
 functions_form(Line, Module, Def, Defmacro, Exports, Body, Deprecated, Struct) ->
   {Spec, Info} = add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct),
-  [{attribute, Line, export, lists:sort([{'__info__', 1} | Exports])}, Spec, Info | Body].
+  [{attribute, Line, export, lists:usort([{'__info__', 1} | Exports])}, Spec, Info | Body].
 
 add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct) ->
+  DefNA = [NA || {NA, _Meta} <- Def],
+  DefmacroNA = [NA || {NA, _Meta} <- Defmacro],
+
   AllowedAttrs = [attributes, compile, functions, macros, md5, exports_md5, module, deprecated, struct],
   AllowedArgs = lists:map(fun(Atom) -> {atom, Line, Atom} end, AllowedAttrs),
-  SortedDef = lists:sort(Def),
-  SortedDefmacro = lists:sort(Defmacro),
 
   Spec =
     {attribute, Line, spec, {{'__info__', 1},
@@ -268,10 +316,10 @@ add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct) ->
   Info =
     {function, 0, '__info__', 1, [
       get_module_info(Module),
-      functions_info(SortedDef),
-      macros_info(SortedDefmacro),
+      functions_info(DefNA),
+      macros_info(DefmacroNA),
       struct_info(Struct),
-      exports_md5_info(Struct, SortedDef, SortedDefmacro),
+      exports_md5_info(Struct, DefNA, DefmacroNA),
       get_module_info(Module, attributes),
       get_module_info(Module, compile),
       get_module_info(Module, md5),
@@ -298,8 +346,7 @@ macros_info(Defmacro) ->
 struct_info(nil) ->
   {clause, 0, [{atom, 0, struct}], [], [{atom, 0, nil}]};
 struct_info(Fields) ->
-  FieldsWithoutDefault = [maps:remove(default, FieldInfo) || FieldInfo <- Fields],
-  {clause, 0, [{atom, 0, struct}], [], [elixir_to_erl(FieldsWithoutDefault)]}.
+  {clause, 0, [{atom, 0, struct}], [], [elixir_to_erl(Fields)]}.
 
 get_module_info(Module, Key) ->
   Call = ?remote(0, erlang, get_module_info, [{atom, 0, Module}, {var, 0, 'Key'}]),
@@ -323,19 +370,34 @@ typespecs_form(Map, TranslatedTypespecs, MacroNames) ->
   Forms2 = callspecs_form(spec, Specs, [], MacroNames, Forms1, Map),
   Forms3 = callspecs_form(callback, AllCallbacks, OptionalCallbacks, MacroCallbackNames, Forms2, Map),
 
-  AllCallbacksWithoutSpecs = lists:usort([
-    {Kind, Name, Arity} || {Kind, {Name, Arity}, _Line, _Spec} <- AllCallbacks
+  AllCallbacksWithoutSpecs = usort_callbacks([
+    {{Kind, Name, Arity}, Meta} || {Kind, {Name, Arity}, Meta, _Spec} <- AllCallbacks
   ]),
 
   {Types, AllCallbacksWithoutSpecs, Forms3}.
 
+usort_callbacks(Callbacks) ->
+  % Sort and deduplicate callbacks. For duplicated callbacks we take
+  % the one with earliest line.
+
+  LineComparator = fun
+    ({Callback1, Meta1}, {Callback1, Meta2}) -> ?line(Meta1) =< ?line(Meta2);
+    ({Callback1, _Meta1}, {Callback2, _Meta2}) -> Callback1 =< Callback2
+  end,
+
+  UniqFun = fun({Callback, _Meta}) -> Callback end,
+
+  lists:uniq(UniqFun, lists:sort(LineComparator, Callbacks)).
+
 %% Types
 
 types_form(Types, Forms) ->
   Fun = fun
-    ({Kind, NameArity, Line, Expr, true}, Acc) ->
+    ({Kind, NameArity, Meta, Expr, true}, Acc) ->
+      Line = ?line(Meta),
       [{attribute, Line, export_type, [NameArity]}, {attribute, Line, Kind, Expr} | Acc];
-    ({Kind, _NameArity, Line, Expr, false}, Acc) ->
+    ({Kind, _NameArity, Meta, Expr, false}, Acc) ->
+      Line = ?line(Meta),
       [{attribute, Line, Kind, Expr} | Acc]
   end,
 
@@ -379,7 +441,9 @@ callspecs_form(Kind, Entries, Optional, Macros, Forms, ModuleMap) ->
   #{unreachable := Unreachable} = ModuleMap,
 
   {SpecsMap, Signatures} =
-    lists:foldl(fun({_, NameArity, Line, Spec}, {Acc, NA}) ->
+    lists:foldl(fun({_, NameArity, Meta, Spec}, {Acc, NA}) ->
+      Line = ?line(Meta),
+
       case Kind of
         spec -> validate_spec_for_existing_function(ModuleMap, NameArity, Line);
         _ -> ok
@@ -418,7 +482,7 @@ callspecs_form(Kind, Entries, Optional, Macros, Forms, ModuleMap) ->
       false ->
         [{attribute, Line, Kind, {Key, lists:reverse(Value)}} | Acc]
     end
-  end, Forms, lists:sort(Signatures)).
+  end, Forms, lists:usort(Signatures)).
 
 spec_for_macro({type, Line, 'bounded_fun', [H | T]}) ->
   {type, Line, 'bounded_fun', [spec_for_macro(H) | T]};
@@ -432,7 +496,7 @@ validate_spec_for_existing_function(ModuleMap, NameAndArity, Line) ->
 
   case lists:keymember(NameAndArity, 1, Defs) of
     true -> ok;
-    false -> file_error(#{line => Line, file => File}, {spec_for_undefined_function, NameAndArity})
+    false -> file_error(#{anno => erl_anno:new(Line), file => File}, {spec_for_undefined_function, NameAndArity})
   end.
 
 % Attributes
@@ -445,7 +509,7 @@ attributes_form(Line, Attributes, Forms) ->
 
 load_form(#{file := File, compile_opts := Opts} = Map, Prefix, Forms, Specs, Chunks) ->
   CompileOpts = extra_chunks_opts(Chunks, debug_opts(Map, Specs, Opts)),
-  {_, Binary} = elixir_erl_compiler:forms(Prefix ++ Specs ++ Forms, File, CompileOpts),
+  {_, Binary} = elixir_erl_compiler:noenv_forms(Prefix ++ Specs ++ Forms, File, CompileOpts),
   Binary.
 
 debug_opts(Map, Specs, Opts) ->
@@ -464,24 +528,32 @@ take_debug_opts(Opts) ->
 extra_chunks_opts([], Opts) -> Opts;
 extra_chunks_opts(Chunks, Opts) -> [{extra_chunks, Chunks} | Opts].
 
-docs_chunk(Set, Module, Line, Def, Defmacro, Types, Callbacks) ->
+docs_chunk(Map, Set, Module, Anno, Def, Defmacro, Types, Callbacks, ChunkOpts) ->
+  #{file := File, attributes := Attributes} = Map,
+
   case elixir_config:get(docs) of
     true ->
-      {ModuleDocLine, ModuleDoc} = get_moduledoc(Line, Set),
+      {ModuleDocLine, ModuleDoc} = get_moduledoc(erl_anno:line(Anno), Set),
       ModuleDocMeta = get_moduledoc_meta(Set),
       FunctionDocs = get_docs(Set, Module, Def, function),
       MacroDocs = get_docs(Set, Module, Defmacro, macro),
       CallbackDocs = get_callback_docs(Set, Callbacks),
       TypeDocs = get_type_docs(Set, Types),
 
+      ModuleMeta = ModuleDocMeta#{
+        source_path => elixir_utils:characters_to_list(File),
+        source_annos => [Anno],
+        behaviours => [Mod || {behaviour, Mod} <- Attributes]
+      },
+
       DocsChunkData = term_to_binary({docs_v1,
         erl_anno:new(ModuleDocLine),
         elixir,
         <<"text/markdown">>,
         ModuleDoc,
-        ModuleDocMeta,
+        ModuleMeta,
         FunctionDocs ++ MacroDocs ++ CallbackDocs ++ TypeDocs
-      }, [deterministic, compressed]),
+      }, [compressed | ChunkOpts]),
 
       [{<<"Docs">>, DocsChunkData}];
 
@@ -521,8 +593,8 @@ get_docs(Set, Module, Definitions, Kind) ->
     maybe_generated(erl_anno:new(Line), Ctx),
     [signature_to_binary(Module, Name, Signature)],
     doc_value(Doc, Name),
-    Meta
-   } || {Name, Arity} <- Definitions,
+    Meta#{source_annos => [?ann(DefinitionMeta)]}
+   } || {{Name, Arity}, DefinitionMeta} <- Definitions,
         {Key, Ctx, Line, Signature, Doc, Meta} <- ets:lookup(Set, {Kind, Name, Arity})].
 
 maybe_generated(Ann, nil) -> Ann;
@@ -533,16 +605,16 @@ get_callback_docs(Set, Callbacks) ->
     erl_anno:new(Line),
     [],
     doc_value(Doc, Name),
-    Meta
-   } || Callback <- Callbacks, {{_, Name, _} = Key, Line, Doc, Meta} <- ets:lookup(Set, Callback)].
+    Meta#{source_annos => [?ann(DefinitionMeta)]}
+   } || {{Kind, Name, Arity}, DefinitionMeta} <- Callbacks, {Key, Line, Doc, Meta} <- ets:lookup(Set, {Kind, Name, Arity})].
 
 get_type_docs(Set, Types) ->
   [{Key,
     erl_anno:new(Line),
     [],
     doc_value(Doc, Name),
-    Meta
-   } || {_Kind, {Name, Arity}, _, _, true} <- Types,
+    Meta#{source_annos => [?ann(DefinitionMeta)]}
+   } || {_Kind, {Name, Arity}, DefinitionMeta, _, true} <- Types,
         {Key, Line, Doc, Meta} <- ets:lookup(Set, {type, Name, Arity})].
 
 signature_to_binary(_Module, Name, _Signature) when Name == '__aliases__'; Name == '__block__' ->
@@ -567,32 +639,33 @@ signature_to_binary(_, Name, Signature) ->
   Doc = 'Elixir.Inspect.Algebra':format('Elixir.Code':quoted_to_algebra(Quoted), infinity),
   'Elixir.IO':iodata_to_binary(Doc).
 
-checker_chunk(#{definitions := Definitions, deprecated := Deprecated, defines_behaviour := DefinesBehaviour}) ->
+checker_chunk(Map, Def, ChunkOpts) ->
+  #{deprecated := Deprecated, defines_behaviour := DefinesBehaviour, signatures := Signatures} = Map,
   DeprecatedMap = maps:from_list(Deprecated),
 
   Exports =
-    lists:foldl(fun({Function, Kind, _Meta, _Clauses}, Acc) ->
-      case Kind of
-        _ when Kind == def orelse Kind == defmacro ->
-          Reason = maps:get(Function, DeprecatedMap, nil),
-          [{Function, #{kind => Kind, deprecated_reason => Reason}} | Acc];
-        _ ->
-          Acc
-      end
-    end, [], Definitions),
+    [begin
+      Signature = maps:get(FA, Signatures, none),
+      Info = case DeprecatedMap of
+        #{FA := Reason} -> #{deprecated => Reason, sig => Signature};
+        #{} -> #{sig => Signature}
+      end,
+      {FA, Info}
+    end || {FA, _Meta} <- prepend_behaviour_info(DefinesBehaviour, Def)],
 
   Contents = #{
-    exports => lists:sort(behaviour_info_exports(DefinesBehaviour) ++ Exports)
+    exports => Exports
   },
 
-  [{<<"ExCk">>, term_to_binary({elixir_checker_v1, Contents}, [deterministic])}].
+  [{<<"ExCk">>, term_to_binary({elixir_checker_v1, Contents}, ChunkOpts)}].
 
-behaviour_info_exports(true) -> [{{behaviour_info, 1}, #{kind => def, deprecated_reason => nil}}];
-behaviour_info_exports(false) -> [].
+prepend_behaviour_info(true, Def) -> [{{behaviour_info, 1}, []} | Def];
+prepend_behaviour_info(false, Def) -> Def.
 
 %% Errors
 
-file_error(#{line := Line, file := File}, Error) ->
+file_error(#{anno := Anno, file := File}, Error) ->
+  Line = erl_anno:line(Anno),
   elixir_errors:file_error([{line, Line}], File, ?MODULE, Error).
 
 format_error({ill_defined_optional_callback, Callback}) ->
