@@ -5,11 +5,13 @@
 
 %% =
 
+expand({'=', Meta, [_, _]} = Expr, S, #{context := match} = E) ->
+  elixir_clauses:parallel_match(Meta, Expr, S, E);
+
 expand({'=', Meta, [Left, Right]}, S, E) ->
   assert_no_guard_scope(Meta, "=", S, E),
   {ERight, SR, ER} = expand(Right, S, E),
-  {ELeft, SL, EL} = elixir_clauses:match(fun expand/3, Left, SR, S, ER),
-  refute_parallel_bitstring_match(ELeft, ERight, E, ?key(E, context) == match),
+  {ELeft, SL, EL} = elixir_clauses:match(fun expand/3, Meta, Left, SR, S, ER),
   {{'=', Meta, [ELeft, ERight]}, SL, EL};
 
 %% Literal operators
@@ -69,12 +71,24 @@ expand({alias, Meta, [Ref, Opts]}, S, E) ->
   {ERef, SR, ER} = expand_without_aliases_report(Ref, S, E),
   {EOpts, ST, ET} = expand_opts(Meta, alias, [as, warn], no_alias_opts(Opts), SR, ER),
 
-  if
-    is_atom(ERef) ->
-      {ERef, ST, alias(Meta, ERef, true, EOpts, ET)};
-    true ->
-      file_error(Meta, E, ?MODULE, {expected_compile_time_module, alias, Ref})
-  end;
+  is_atom(ERef) orelse
+    file_error(Meta, E, ?MODULE, {expected_compile_time_module, alias, Ref}),
+
+  {ok, New, EQ} = alias(Meta, ERef, true, EOpts, ET),
+
+  Quoted =
+    case (New /= false) andalso should_warn(Meta, EOpts, EQ) of
+      false ->
+        ERef;
+
+      Pid when ?key(EQ, function) /= nil ->
+        ?tracker:warn_alias(Pid, Meta, New, ERef);
+
+      Pid ->
+        {{'.', Meta, [?tracker, warn_alias]}, Meta, [Pid, Meta, New, ERef]}
+    end,
+
+  {Quoted, ST, EQ};
 
 expand({require, Meta, [Ref]}, S, E) ->
   expand({require, Meta, [Ref, []]}, S, E);
@@ -89,19 +103,21 @@ expand({require, Meta, [Ref, Opts]}, S, E) ->
   %% module in context modules.
   case lists:keyfind(defined, 1, Meta) of
     {defined, Mod} when is_atom(Mod) ->
-      EA = ET#{context_modules := [Mod | ?key(ET, context_modules)]},
+      Mods = [Mod | ?key(ET, context_modules)],
+      {ok, _, EU} = alias(Meta, ERef, false, EOpts, ET#{context_modules := Mods}),
 
       SU = case E of
         #{function := nil} -> ST;
         _ -> ST#elixir_ex{runtime_modules=[Mod | ST#elixir_ex.runtime_modules]}
       end,
 
-      {ERef, SU, alias(Meta, ERef, false, EOpts, EA)};
+      {ERef, SU, EU};
 
     false when is_atom(ERef) ->
       elixir_aliases:ensure_loaded(Meta, ERef, ET),
       RE = elixir_aliases:require(Meta, ERef, EOpts, ET, true),
-      {ERef, ST, alias(Meta, ERef, false, EOpts, RE)};
+      {ok, _, EU} = alias(Meta, ERef, false, EOpts, RE),
+      {ERef, ST, EU};
 
     false ->
       file_error(Meta, E, ?MODULE, {expected_compile_time_module, require, Ref})
@@ -115,16 +131,41 @@ expand({import, Meta, [Ref, Opts]}, S, E) ->
   {ERef, SR, ER} = expand_without_aliases_report(Ref, S, E),
   {EOpts, ST, ET} = expand_opts(Meta, import, [only, except, warn], Opts, SR, ER),
 
-  if
-    is_atom(ERef) ->
-      elixir_aliases:ensure_loaded(Meta, ERef, ET),
+  is_atom(ERef) orelse
+    file_error(Meta, E, ?MODULE, {expected_compile_time_module, import, Ref}),
 
-      case elixir_import:import(Meta, ERef, EOpts, ET, true, true) of
-        {ok, EI} -> {ERef, ST, EI};
-        {error, Reason} -> elixir_errors:file_error(Meta, E, elixir_import, Reason)
-      end;
-    true ->
-      file_error(Meta, E, ?MODULE, {expected_compile_time_module, import, Ref})
+  elixir_aliases:ensure_loaded(Meta, ERef, ET),
+
+  case elixir_import:import(Meta, ERef, EOpts, ET, true, true) of
+    {ok, Imported, EI} ->
+      Quoted =
+        case Imported andalso should_warn(Meta, EOpts, ET) of
+          false ->
+            ERef;
+
+          Pid ->
+            Only =
+              case lists:keyfind(only, 1, EOpts) of
+                {only, List} when is_list(List) -> List;
+                _ -> []
+              end,
+
+            %% If we are outside a function, we turn on the warnings at execution time.
+            case ET of
+              #{function := nil} ->
+                ?tracker:add_import(Pid, ERef, Only, Meta, false),
+                {{'.', Meta, [?tracker, warn_import]}, Meta, [Pid, ERef]};
+
+              #{} ->
+                ?tracker:add_import(Pid, ERef, Only, Meta, true),
+                ERef
+            end
+        end,
+
+        {Quoted, ST, EI};
+
+    {error, Reason} ->
+      elixir_errors:file_error(Meta, E, elixir_import, Reason)
   end;
 
 %% Compilation environment macros
@@ -252,7 +293,7 @@ expand({'&', Meta, [{'/', ArityMeta, [{super, SuperMeta, Context}, Arity]} = Exp
 
   case resolve_super(Meta, Arity, E) of
     {Kind, Name, _} when Kind == def; Kind == defp ->
-      {{'&', Meta, [{'/', ArityMeta, [{Name, SuperMeta, Context}, Arity]}]}, S, E};
+      expand({'&', Meta, [{'/', ArityMeta, [{Name, SuperMeta, Context}, Arity]}]}, S, E);
     _ ->
       expand_fn_capture(Meta, Expr, S, E)
   end;
@@ -302,7 +343,9 @@ expand({with, Meta, [_ | _] = Args}, S, E) ->
 
 expand({super, Meta, Args}, S, E) when is_list(Args) ->
   assert_no_match_or_guard_scope(Meta, "super", S, E),
-  {Kind, Name, _} = resolve_super(Meta, length(Args), E),
+  Arity = length(Args),
+  {Kind, Name, _} = resolve_super(Meta, Arity, E),
+  elixir_env:trace({local_function, Meta, Name, Arity}, E),
   {EArgs, SA, EA} = expand_args(Args, S, E),
   {{super, [{super, {Kind, Name}} | Meta], EArgs}, SA, EA};
 
@@ -329,7 +372,7 @@ expand({'_', Meta, Kind} = Var, S, #{context := Context} = E) when is_atom(Kind)
 
 expand({Name, Meta, Kind}, S, #{context := match} = E) when is_atom(Name), is_atom(Kind) ->
   #elixir_ex{
-    prematch={_, PrematchVersion, _},
+    prematch={_, _, PrematchVersion},
     unused={Unused, Version},
     vars={Read, Write}
   } = S,
@@ -341,8 +384,9 @@ expand({Name, Meta, Kind}, S, #{context := match} = E) when is_atom(Name), is_at
     #{Pair := VarVersion} when VarVersion >= PrematchVersion ->
       maybe_warn_underscored_var_repeat(Meta, Name, Kind, E),
       NewUnused = var_used(Pair, Meta, VarVersion, Unused),
+      NewWrite = (Write /= false) andalso Write#{Pair => Version},
       Var = {Name, [{version, VarVersion} | Meta], Kind},
-      {Var, S#elixir_ex{unused={NewUnused, Version}}, E};
+      {Var, S#elixir_ex{vars={Read, NewWrite}, unused={NewUnused, Version}}, E};
 
     %% Variable is being overridden now
     #{Pair := _} ->
@@ -369,17 +413,20 @@ expand({Name, Meta, Kind}, S, E) when is_atom(Name), is_atom(Kind) ->
     case Read of
       #{Pair := CurrentVersion} ->
         case Prematch of
-          {Pre, _Counter, {bitsize, Original}} ->
+          {Pre, _Cycle, {bitsize, Original}} ->
             if
               map_get(Pair, Pre) /= CurrentVersion ->
                 {ok, CurrentVersion};
+
               is_map_key(Pair, Pre) ->
                 %% TODO: Enable this warning on Elixir v1.19
                 %% TODO: Remove me on Elixir 2.0
                 %% elixir_errors:file_warn(Meta, E, ?MODULE, {unpinned_bitsize_var, Name, Kind}),
                 {ok, CurrentVersion};
+
               not is_map_key(Pair, Original) ->
                 {ok, CurrentVersion};
+
               true ->
                 raise
             end;
@@ -389,7 +436,12 @@ expand({Name, Meta, Kind}, S, E) when is_atom(Name), is_atom(Kind) ->
         end;
 
       _ ->
-        Prematch
+        case E of
+          #{context := guard} -> raise;
+          #{} when S#elixir_ex.prematch =:= pin -> pin;
+          %% TODO: Remove fallback on on_undefined_variable
+          _ -> elixir_config:get(on_undefined_variable)
+        end
     end,
 
   case Result of
@@ -417,7 +469,7 @@ expand({Name, Meta, Kind}, S, E) when is_atom(Name), is_atom(Kind) ->
           function_error(Meta, E, ?MODULE, {undefined_var_pin, Name, Kind}),
           {{Name, Meta, Kind}, S, E};
 
-        _ ->
+        _ when Error == raise ->
           SpanMeta =  elixir_env:calculate_span(Meta, Name),
           function_error(SpanMeta, E, ?MODULE, {undefined_var, Name, Kind}),
           {{Name, SpanMeta, Kind}, S, E}
@@ -429,18 +481,22 @@ expand({Name, Meta, Kind}, S, E) when is_atom(Name), is_atom(Kind) ->
 expand({Atom, Meta, Args}, S, E) when is_atom(Atom), is_list(Meta), is_list(Args) ->
   assert_no_ambiguous_op(Atom, Meta, Args, S, E),
 
-  elixir_dispatch:dispatch_import(Meta, Atom, Args, S, E, fun() ->
-    expand_local(Meta, Atom, Args, S, E)
+  elixir_dispatch:dispatch_import(Meta, Atom, Args, S, E, fun
+    ({AR, AF}) ->
+      expand_remote(AR, Meta, AF, Meta, Args, S, elixir_env:prepare_write(S, E), E);
+
+    (local) ->
+      expand_local(Meta, Atom, Args, S, E)
   end);
 
 %% Remote calls
 
 expand({{'.', DotMeta, [Left, Right]}, Meta, Args}, S, E)
     when (is_tuple(Left) orelse is_atom(Left)), is_atom(Right), is_list(Meta), is_list(Args) ->
-  {ELeft, SL, EL} = expand(Left, elixir_env:prepare_write(S), E),
+  {ELeft, SL, EL} = expand(Left, elixir_env:prepare_write(S, E), E),
 
-  elixir_dispatch:dispatch_require(Meta, ELeft, Right, Args, S, EL, fun(AR, AF, AA) ->
-    expand_remote(AR, DotMeta, AF, Meta, AA, S, SL, EL)
+  elixir_dispatch:dispatch_require(Meta, ELeft, Right, Args, S, EL, fun(AR, AF) ->
+    expand_remote(AR, DotMeta, AF, Meta, Args, S, SL, EL)
   end);
 
 %% Anonymous calls
@@ -448,7 +504,6 @@ expand({{'.', DotMeta, [Left, Right]}, Meta, Args}, S, E)
 expand({{'.', DotMeta, [Expr]}, Meta, Args}, S, E) when is_list(Args) ->
   assert_no_match_or_guard_scope(Meta, "anonymous call", S, E),
   {[EExpr | EArgs], SA, EA} = expand_args([Expr | Args], S, E),
-  is_atom(EExpr) andalso function_error(Meta, E, ?MODULE, {invalid_function_call, EExpr}),
   {{{'.', DotMeta, [EExpr]}, Meta, EArgs}, SA, EA};
 
 %% Invalid calls
@@ -471,9 +526,16 @@ expand(List, S, E) when is_list(List) ->
 
   {EArgs, elixir_env:close_write(SE, S), EE};
 
+expand(Zero, S, #{context := match} = E) when is_float(Zero), Zero == 0.0 ->
+  elixir_errors:file_warn([], E, ?MODULE, invalid_match_on_zero_float),
+  {Zero, S, E};
+
+expand(Other, S, E) when is_number(Other); is_atom(Other); is_binary(Other) ->
+  {Other, S, E};
+
 expand(Function, S, E) when is_function(Function) ->
   case (erlang:fun_info(Function, type) == {type, external}) andalso
-       (erlang:fun_info(Function, env) == {env, []}) of
+        (erlang:fun_info(Function, env) == {env, []}) of
     true ->
       {elixir_quote:fun_to_quoted(Function), S, E};
     false ->
@@ -490,13 +552,6 @@ expand(Pid, S, E) when is_pid(Pid) ->
       {Pid, S, E}
   end;
 
-expand(Zero, S, #{context := match} = E) when is_float(Zero), Zero == 0.0 ->
-  elixir_errors:file_warn([], E, ?MODULE, invalid_match_on_zero_float),
-  {Zero, S, E};
-
-expand(Other, S, E) when is_number(Other); is_atom(Other); is_binary(Other) ->
-  {Other, S, E};
-
 expand(Other, _S, E) ->
   file_error([{line, 0}], ?key(E, file), ?MODULE, {invalid_quoted_expr, Other}).
 
@@ -697,9 +752,6 @@ maybe_warn_deprecated_super_in_gen_server_callback(Meta, Function, SuperMeta, E)
       ok
   end.
 
-context_info(Kind) when Kind == nil; is_integer(Kind) -> "";
-context_info(Kind) -> io_lib:format(" (context ~ts)", [elixir_aliases:inspect(Kind)]).
-
 should_warn(Meta) ->
   lists:keyfind(generated, 1, Meta) /= {generated, true}.
 
@@ -709,15 +761,9 @@ expand_case(Meta, Expr, Opts, S, E) ->
   {EExpr, SE, EE} = expand(Expr, S, E),
 
   ROpts =
-    case proplists:get_value(optimize_boolean, Meta, false) of
-      true ->
-        case elixir_utils:returns_boolean(EExpr) of
-          true -> rewrite_case_clauses(Opts);
-          false -> generated_case_clauses(Opts)
-        end;
-
-      false ->
-        Opts
+    case lists:member({optimize_boolean, true}, Meta) andalso elixir_utils:returns_boolean(EExpr) of
+      true -> rewrite_case_clauses(Opts);
+      false -> Opts
     end,
 
   {EOpts, SO, EO} = elixir_clauses:'case'(Meta, ROpts, SE, EE),
@@ -741,19 +787,15 @@ rewrite_case_clauses([{do, [
 ]}]) ->
   rewrite_case_clauses(FalseMeta, FalseExpr, TrueMeta, TrueExpr);
 
-rewrite_case_clauses(Other) ->
-  generated_case_clauses(Other).
+rewrite_case_clauses(Opts) ->
+  Opts.
 
 rewrite_case_clauses(FalseMeta, FalseExpr, TrueMeta, TrueExpr) ->
   [{do, [
-    {'->', ?generated(FalseMeta), [[false], FalseExpr]},
-    {'->', ?generated(TrueMeta), [[true], TrueExpr]}
+    {'->', FalseMeta, [[false], FalseExpr]},
+    {'->', TrueMeta, [[true], TrueExpr]}
   ]}].
 
-generated_case_clauses([{do, Clauses}]) ->
-  RClauses = [{'->', ?generated(Meta), Args} || {'->', Meta, Args} <- Clauses],
-  [{do, RClauses}].
-
 %% Comprehensions
 
 expand_for({for, Meta, [_ | _] = Args}, S, E, Return) ->
@@ -809,11 +851,14 @@ expand_for_do_block(_Meta, Expr, S, E, false) ->
   expand(Expr, S, E);
 expand_for_do_block(Meta, [{'->', _, _} | _] = Clauses, S, E, {reduce, _}) ->
   Transformer = fun
+    ({_, _, [[{'when', _, [_, _, _ | _]}], _]}, _) ->
+      file_error(Meta, E, ?MODULE, for_with_reduce_bad_block);
+
     ({_, _, [[_], _]} = Clause, SA) ->
       SReset = elixir_env:reset_unused_vars(SA),
 
       {EClause, SAcc, EAcc} =
-        elixir_clauses:clause(Meta, fn, fun elixir_clauses:head/3, Clause, SReset, E),
+        elixir_clauses:clause(Meta, fn, fun elixir_clauses:head/4, Clause, SReset, E),
 
       {EClause, elixir_env:merge_and_check_unused_vars(SAcc, SA, EAcc)};
 
@@ -844,25 +889,8 @@ assert_no_ambiguous_op(Name, Meta, [Arg], S, E) ->
 assert_no_ambiguous_op(_Atom, _Meta, _Args, _S, _E) ->
   ok.
 
-assert_no_clauses(_Name, _Meta, [], _E) ->
-  ok;
-assert_no_clauses(Name, Meta, Args, E) ->
-  assert_arg_with_no_clauses(Name, Meta, lists:last(Args), E).
-
-assert_arg_with_no_clauses(Name, Meta, [{Key, Value} | Rest], E) when is_atom(Key) ->
-  case Value of
-    [{'->', _, _} | _] ->
-      file_error(Meta, E, ?MODULE, {invalid_clauses, Name});
-    _ ->
-      assert_arg_with_no_clauses(Name, Meta, Rest, E)
-  end;
-assert_arg_with_no_clauses(_Name, _Meta, _Arg, _E) ->
-  ok.
-
-expand_local(Meta, Name, Args, S, #{module := Module, function := Function, context := Context} = E)
+expand_local(Meta, Name, Args, S, #{function := Function, context := Context} = E)
     when Function /= nil ->
-  assert_no_clauses(Name, Meta, Args, E),
-
   %% In case we have the wrong context, we log a module error
   %% so we can print multiple entries at the same time.
   case Context of
@@ -870,12 +898,10 @@ expand_local(Meta, Name, Args, S, #{module := Module, function := Function, cont
       module_error(Meta, E, ?MODULE, {invalid_local_invocation, "match", {Name, Meta, Args}});
 
     guard ->
-      module_error(Meta, E, ?MODULE, {invalid_local_invocation, guard_context(S), {Name, Meta, Args}});
+      module_error(Meta, E, ?MODULE, {invalid_local_invocation, elixir_utils:guard_info(S), {Name, Meta, Args}});
 
     nil ->
-      Arity = length(Args),
-      elixir_env:trace({local_function, Meta, Name, Arity}, E),
-      elixir_locals:record_local({Name, Arity}, Module, Function, Meta, false)
+      elixir_env:trace({local_function, Meta, Name, length(Args)}, E)
   end,
 
   {EArgs, SA, EA} = expand_args(Args, S, E),
@@ -887,26 +913,30 @@ expand_local(Meta, Name, Args, _S, #{function := nil} = E) ->
 
 expand_remote(Receiver, DotMeta, Right, Meta, Args, S, SL, #{context := Context} = E)
     when is_atom(Receiver) or is_tuple(Receiver) ->
-  assert_no_clauses(Right, Meta, Args, E),
-
   if
     Context =:= guard, is_tuple(Receiver) ->
       (lists:keyfind(no_parens, 1, Meta) /= {no_parens, true}) andalso
-        function_error(Meta, E, ?MODULE, {parens_map_lookup, Receiver, Right, guard_context(S)}),
+        function_error(Meta, E, ?MODULE, {parens_map_lookup, Receiver, Right, elixir_utils:guard_info(S)}),
 
       {{{'.', DotMeta, [Receiver, Right]}, Meta, []}, SL, E};
 
-    true ->
+    Context =:= nil ->
       AttachedMeta = attach_runtime_module(Receiver, Meta, S, E),
       {EArgs, {SA, _}, EA} = mapfold(fun expand_arg/3, {SL, S}, E, Args),
+      Rewritten = elixir_rewrite:rewrite(Receiver, DotMeta, Right, AttachedMeta, EArgs),
+      {Rewritten, elixir_env:close_write(SA, S), EA};
 
-      case rewrite(Context, Receiver, DotMeta, Right, AttachedMeta, EArgs, S) of
-        {ok, Rewritten} ->
-          maybe_warn_comparison(Rewritten, Args, E),
-          {Rewritten, elixir_env:close_write(SA, S), EA};
+    true ->
+      case {Receiver, Right, Args} of
+        {erlang, '+', [Arg]} when is_number(Arg) -> {+Arg, SL, E};
+        {erlang, '-', [Arg]} when is_number(Arg) -> {-Arg, SL, E};
+        _ ->
+          {EArgs, SA, EA} = mapfold(fun expand/3, SL, E, Args),
 
-        {error, Error} ->
-          file_error(Meta, E, elixir_rewrite, Error)
+          case elixir_rewrite:Context(Receiver, DotMeta, Right, Meta, EArgs, S) of
+            {ok, Rewritten} -> {Rewritten, SA, EA};
+            {error, Error} -> file_error(Meta, E, elixir_rewrite, Error)
+          end
       end
   end;
 expand_remote(Receiver, DotMeta, Right, Meta, Args, _, _, E) ->
@@ -919,42 +949,6 @@ attach_runtime_module(Receiver, Meta, S, _E) ->
     false -> Meta
   end.
 
-% Signed numbers can be rewritten no matter the context
-rewrite(_, erlang, _, '+', _, [Arg], _S) when is_number(Arg) -> {ok, Arg};
-rewrite(_, erlang, _, '-', _, [Arg], _S) when is_number(Arg) -> {ok, -Arg};
-rewrite(match, Receiver, DotMeta, Right, Meta, EArgs, _S) ->
-  elixir_rewrite:match_rewrite(Receiver, DotMeta, Right, Meta, EArgs);
-rewrite(guard, Receiver, DotMeta, Right, Meta, EArgs, S) ->
-  elixir_rewrite:guard_rewrite(Receiver, DotMeta, Right, Meta, EArgs, guard_context(S));
-rewrite(_, Receiver, DotMeta, Right, Meta, EArgs, _S) ->
-  {ok, elixir_rewrite:rewrite(Receiver, DotMeta, Right, Meta, EArgs)}.
-
-maybe_warn_comparison({{'.', _, [erlang, Op]}, Meta, [ELeft, ERight]}, [Left, Right], E)
-    when Op =:= '>'; Op =:= '<'; Op =:= '=<'; Op =:= '>='; Op =:= min; Op =:= max ->
-
-    case is_nested_comparison(Op, ELeft, ERight, Left, Right) of
-      false -> ok;
-      CompExpr ->
-        elixir_errors:file_warn(Meta, E, ?MODULE, {nested_comparison, CompExpr})
-    end;
-maybe_warn_comparison(_, _, _) ->
-  ok.
-
-is_nested_comparison(Op, ELeft, ERight, Left, Right) ->
-  NestedExpr = {elixir_utils:erlang_comparison_op_to_elixir(Op), [], [Left, Right]},
-  case is_comparison_expression(ELeft) of
-    true ->
-      NestedExpr;
-    false ->
-      case is_comparison_expression(ERight) of
-        true -> NestedExpr;
-        false -> false
-      end
-  end.
-is_comparison_expression({{'.',_,[erlang, Op]},_,_})
-  when Op =:= '>'; Op =:= '<'; Op =:= '=<'; Op =:= '>=' -> true;
-is_comparison_expression(_Other) -> false.
-
 %% Lexical helpers
 
 expand_opts(Meta, Kind, Allowed, Opts, S, E) ->
@@ -982,11 +976,24 @@ no_alias_expansion({'__aliases__', _, [H | T]}) when is_atom(H) ->
 no_alias_expansion(Other) ->
   Other.
 
+should_warn(_Meta, _Opts, #{lexical_tracker := nil}) ->
+  false;
+should_warn(Meta, Opts, #{lexical_tracker := Pid}) ->
+  case lists:keyfind(warn, 1, Opts) of
+    {warn, false} -> false;
+    {warn, true} -> Pid;
+    false ->
+      case lists:keymember(context, 1, Meta) of
+        true -> false;
+        false -> Pid
+      end
+  end.
+
 %% Aliases
 
 alias(Meta, Ref, IncludeByDefault, Opts, E) ->
   case elixir_aliases:alias(Meta, Ref, IncludeByDefault, Opts, E, true) of
-    {ok, EA} -> EA;
+    {ok, New, EA} -> {ok, New, EA};
     {error, Reason} -> elixir_errors:file_error(Meta, E, elixir_aliases, Reason)
   end.
 
@@ -1026,7 +1033,7 @@ expand_aliases({'__aliases__', Meta, List} = Alias, S, E, Report) ->
 expand_for_generator({'<-', Meta, [Left, Right]}, S, E) ->
   {ERight, SR, ER} = expand(Right, S, E),
   SM = elixir_env:reset_read(SR, S),
-  {[ELeft], SL, EL} = elixir_clauses:head([Left], SM, ER),
+  {[ELeft], SL, EL} = elixir_clauses:head(Meta, [Left], SM, ER),
   {{'<-', Meta, [ELeft, ERight]}, SL, EL};
 expand_for_generator({'<<>>', Meta, Args} = X, S, E) when is_list(Args) ->
   case elixir_utils:split_last(Args) of
@@ -1035,7 +1042,7 @@ expand_for_generator({'<<>>', Meta, Args} = X, S, E) when is_list(Args) ->
       SM = elixir_env:reset_read(SR, S),
       {ELeft, SL, EL} = elixir_clauses:match(fun(BArg, BS, BE) ->
         elixir_bitstring:expand(Meta, BArg, BS, BE, true)
-      end, LeftStart ++ [LeftEnd], SM, SM, ER),
+      end, Meta, LeftStart ++ [LeftEnd], SM, SM, ER),
       {{'<<>>', Meta, [{'<-', OpMeta, [ELeft, ERight]}]}, SL, EL};
     _ ->
       expand(X, S, E)
@@ -1052,45 +1059,6 @@ assert_generator_start(Meta, _, E) ->
 
 %% Assertions
 
-refute_parallel_bitstring_match({'<<>>', _, _}, {'<<>>', Meta, _} = Arg, E, true) ->
-  file_error(Meta, E, ?MODULE, {parallel_bitstring_match, Arg});
-refute_parallel_bitstring_match(Left, {'=', _Meta, [MatchLeft, MatchRight]}, E, Parallel) ->
-  refute_parallel_bitstring_match(Left, MatchLeft, E, true),
-  refute_parallel_bitstring_match(Left, MatchRight, E, Parallel);
-refute_parallel_bitstring_match([_ | _] = Left, [_ | _] = Right, E, Parallel) ->
-  refute_parallel_bitstring_match_each(Left, Right, E, Parallel);
-refute_parallel_bitstring_match({Left1, Left2}, {Right1, Right2}, E, Parallel) ->
-  refute_parallel_bitstring_match_each([Left1, Left2], [Right1, Right2], E, Parallel);
-refute_parallel_bitstring_match({'{}', _, Args1}, {'{}', _, Args2}, E, Parallel) ->
-  refute_parallel_bitstring_match_each(Args1, Args2, E, Parallel);
-refute_parallel_bitstring_match({'%{}', _, Args1}, {'%{}', _, Args2}, E, Parallel) ->
-  refute_parallel_bitstring_match_map_field(lists:sort(Args1), lists:sort(Args2), E, Parallel);
-refute_parallel_bitstring_match({'%', _, [_, Args]}, Right, E, Parallel) ->
-  refute_parallel_bitstring_match(Args, Right, E, Parallel);
-refute_parallel_bitstring_match(Left, {'%', _, [_, Args]}, E, Parallel) ->
-  refute_parallel_bitstring_match(Left, Args, E, Parallel);
-refute_parallel_bitstring_match(_Left, _Right, _E, _Parallel) ->
-  ok.
-
-refute_parallel_bitstring_match_each([Arg1 | Rest1], [Arg2 | Rest2], E, Parallel) ->
-  refute_parallel_bitstring_match(Arg1, Arg2, E, Parallel),
-    refute_parallel_bitstring_match_each(Rest1, Rest2, E, Parallel);
-refute_parallel_bitstring_match_each(_List1, _List2, _E, _Parallel) ->
-  ok.
-
-refute_parallel_bitstring_match_map_field([{Key, Val1} | Rest1], [{Key, Val2} | Rest2], E, Parallel) ->
-  refute_parallel_bitstring_match(Val1, Val2, E, Parallel),
-  refute_parallel_bitstring_match_map_field(Rest1, Rest2, E, Parallel);
-refute_parallel_bitstring_match_map_field([Field1 | Rest1] = Args1, [Field2 | Rest2] = Args2, E, Parallel) ->
-  case Field1 > Field2 of
-    true ->
-      refute_parallel_bitstring_match_map_field(Args1, Rest2, E, Parallel);
-    false ->
-      refute_parallel_bitstring_match_map_field(Rest1, Args2, E, Parallel)
-  end;
-refute_parallel_bitstring_match_map_field(_Args1, _Args2, _E, _Parallel) ->
-  ok.
-
 assert_module_scope(Meta, Kind, #{module := nil, file := File}) ->
   file_error(Meta, File, ?MODULE, {invalid_expr_in_scope, "module", Kind});
 assert_module_scope(_Meta, _Kind, #{module:=Module}) -> Module.
@@ -1107,8 +1075,8 @@ assert_no_match_scope(Meta, Kind, #{context := match, file := File}) ->
 assert_no_match_scope(_Meta, _Kind, _E) -> ok.
 assert_no_guard_scope(Meta, Kind, S, #{context := guard, file := File}) ->
   Key =
-    case S#elixir_ex.prematch of
-      {_, _, {bitsize, _}}  -> invalid_expr_in_bitsize;
+    case S of
+      #elixir_ex{prematch={_, _, {bitsize, _}}}  -> invalid_expr_in_bitsize;
       _ -> invalid_expr_in_guard
     end,
   file_error(Meta, File, ?MODULE, {Key, Kind});
@@ -1132,9 +1100,6 @@ assert_no_underscore_clause_in_cond(_Other, _E) ->
 
 %% Errors
 
-guard_context(#elixir_ex{prematch={_, _, {bitsize, _}}}) -> "bitstring size specifier";
-guard_context(_) -> "guard".
-
 format_error(invalid_match_on_zero_float) ->
   "pattern matching on 0.0 is equivalent to matching only on +0.0 from Erlang/OTP 27+. Instead you must match on +0.0 or -0.0";
 format_error({useless_literal, Term}) ->
@@ -1210,10 +1175,10 @@ format_error({pin_outside_of_match, Arg}) ->
 format_error(unbound_underscore) ->
   "invalid use of _. _ can only be used inside patterns to ignore values and cannot be used in expressions. Make sure you are inside a pattern or change it accordingly";
 format_error({undefined_var, Name, Kind}) ->
-  io_lib:format("undefined variable \"~ts\"~ts", [Name, context_info(Kind)]);
+  io_lib:format("undefined variable ~ts", [elixir_utils:var_info(Name, Kind)]);
 format_error({undefined_var_pin, Name, Kind}) ->
-  Message = "undefined variable ^~ts. No variable \"~ts\"~ts has been defined before the current pattern",
-  io_lib:format(Message, [Name, Name, context_info(Kind)]);
+  Message = "undefined variable ^~ts. No variable ~ts has been defined before the current pattern",
+  io_lib:format(Message, [Name, elixir_utils:var_info(Name, Kind)]);
 format_error(underscore_in_cond) ->
   "invalid use of _ inside \"cond\". If you want the last clause to always match, "
     "you probably meant to use: true ->";
@@ -1256,8 +1221,6 @@ format_error({invalid_clauses, Name}) ->
     "the function \"~ts\" cannot handle clauses with the -> operator because it is not a macro. "
     "Please make sure you are invoking the proper name and that it is a macro",
   io_lib:format(Message, [Name]);
-format_error({invalid_function_call, Expr}) ->
-  io_lib:format("invalid function call :~ts.()", [Expr]);
 format_error({invalid_call, Call}) ->
   io_lib:format("invalid call ~ts", ['Elixir.Macro':to_string(Call)]);
 format_error({invalid_quoted_expr, Expr}) ->
@@ -1284,15 +1247,15 @@ format_error({options_are_not_keyword, Kind, Opts}) ->
 format_error({undefined_function, Name, Args}) ->
   io_lib:format("undefined function ~ts/~B (there is no such import)", [Name, length(Args)]);
 format_error({unpinned_bitsize_var, Name, Kind}) ->
-  io_lib:format("the variable \"~ts\"~ts is accessed inside size(...) of a bitstring "
+  io_lib:format("the variable ~ts is accessed inside size(...) of a bitstring "
                 "but it was defined outside of the match. You must precede it with the "
-                "pin operator", [Name, context_info(Kind)]);
+                "pin operator", [elixir_utils:var_info(Name, Kind)]);
 format_error({underscored_var_repeat, Name, Kind}) ->
-  io_lib:format("the underscored variable \"~ts\"~ts appears more than once in a "
+  io_lib:format("the underscored variable ~ts appears more than once in a "
                 "match. This means the pattern will only match if all \"~ts\" bind "
                 "to the same value. If this is the intended behaviour, please "
                 "remove the leading underscore from the variable name, otherwise "
-                "give the variables different names", [Name, context_info(Kind), Name]);
+                "give the variables different names", [elixir_utils:var_info(Name, Kind), Name]);
 format_error({underscored_var_access, Name}) ->
   io_lib:format("the underscored variable \"~ts\" is used after being set. "
                 "A leading underscore indicates that the value of the variable "
@@ -1324,8 +1287,4 @@ format_error({parens_map_lookup, Map, Field, Context}) ->
 format_error({super_in_genserver, {Name, Arity}}) ->
   io_lib:format("calling super for GenServer callback ~ts/~B is deprecated", [Name, Arity]);
 format_error('__cursor__') ->
-  "reserved special form __cursor__ cannot be expanded, it is used exclusively to annotate ASTs";
-format_error({parallel_bitstring_match, Expr}) ->
-  Message =
-    "binary patterns cannot be matched in parallel using \"=\", excess pattern: ~ts",
-  io_lib:format(Message, ['Elixir.Macro':to_string(Expr)]).
+  "reserved special form __cursor__ cannot be expanded, it is used exclusively to annotate ASTs".
\ No newline at end of file
