@@ -1,9 +1,10 @@
 -module(elixir_module).
 -export([file/1, data_tables/1, is_open/1, mode/1, delete_definition_attributes/6,
-         compile/5, expand_callback/6, format_error/1, compiler_modules/0,
-         write_cache/3, read_cache/2, next_counter/1, taint/1]).
+         compile/6, expand_callback/6, format_error/1, compiler_modules/0,
+         write_cache/3, read_cache/2, next_counter/1, taint/1, cache_env/1, get_cached_env/1]).
 -include("elixir.hrl").
 -define(counter_attr, {elixir, counter}).
+-define(cache_key, {elixir, cache_env}).
 
 %% Stores modules currently being defined by the compiler
 
@@ -71,11 +72,34 @@ taint(Module) ->
     _:_ -> false
   end.
 
+cache_env(#{line := Line, module := Module} = E) ->
+  {Set, _} = data_tables(Module),
+  Cache = elixir_env:reset_vars(E#{line := nil}),
+  PrevKey = ets:lookup_element(Set, ?cache_key, 2),
+
+  Pos =
+    case ets:lookup(Set, {cache_env, PrevKey}) of
+      [{_, Cache}] ->
+        PrevKey;
+      _ ->
+        NewKey = PrevKey + 1,
+        ets:insert(Set, [{{cache_env, NewKey}, Cache}, {?cache_key, NewKey}]),
+        NewKey
+    end,
+
+  {Module, {Line, Pos}}.
+
+get_cached_env({Module, {Line, Pos}}) ->
+  {Set, _} = data_tables(Module),
+  (ets:lookup_element(Set, {cache_env, Pos}, 2))#{line := Line};
+get_cached_env(Env) ->
+  Env.
+
 %% Compilation hook
 
-compile(Module, Block, Vars, Prune, Env) ->
+compile(Meta, Module, Block, Vars, Prune, Env) ->
   ModuleAsCharlist = validate_module_name(Module),
-  #{line := Line, function := Function, versioned_vars := OldVerVars} = Env,
+  #{function := Function, versioned_vars := OldVerVars} = Env,
 
   {VerVars, _} =
     lists:mapfoldl(fun({Var, _}, I) -> {{Var, I}, I + 1} end, 0, maps:to_list(OldVerVars)),
@@ -92,11 +116,11 @@ compile(Module, Block, Vars, Prune, Env) ->
     #{lexical_tracker := nil} ->
       elixir_lexical:run(
         MaybeLexEnv,
-        fun(LexEnv) -> compile(Line, Module, ModuleAsCharlist, Block, Vars, Prune, LexEnv) end,
+        fun(LexEnv) -> compile(Meta, Module, ModuleAsCharlist, Block, Vars, Prune, LexEnv) end,
         fun(_LexEnv) -> ok end
       );
     _ ->
-      compile(Line, Module, ModuleAsCharlist, Block, Vars, Prune, MaybeLexEnv)
+      compile(Meta, Module, ModuleAsCharlist, Block, Vars, Prune, MaybeLexEnv)
   end.
 
 validate_module_name(Module) when Module == nil; is_boolean(Module); not is_atom(Module) ->
@@ -115,7 +139,10 @@ invalid_module_name(Module) ->
       ('Elixir.Kernel':inspect(Module))/binary>>
   )).
 
-compile(Line, Module, ModuleAsCharlist, Block, Vars, Prune, E) ->
+compile(Meta, Module, ModuleAsCharlist, Block, Vars, Prune, E) ->
+  Anno = ?ann(Meta),
+  Line = erl_anno:line(Anno),
+
   File = ?key(E, file),
   check_module_availability(Module, Line, E),
   elixir_env:trace(defmodule, E),
@@ -128,6 +155,7 @@ compile(Line, Module, ModuleAsCharlist, Block, Vars, Prune, E) ->
     put_compiler_modules([Module | CompilerModules]),
     {Result, ModuleE, CallbackE} = eval_form(Line, Module, DataBag, Block, Vars, Prune, E),
     CheckerInfo = checker_info(),
+    BeamLocation = beam_location(ModuleAsCharlist),
 
     {Binary, PersistedAttributes, Autoload} =
       elixir_erl_compiler:spawn(fun() ->
@@ -136,70 +164,69 @@ compile(Line, Module, ModuleAsCharlist, Block, Vars, Prune, E) ->
         {AllDefinitions, Private} = elixir_def:fetch_definitions(Module, E),
 
         OnLoadAttribute = lists:keyfind(on_load, 1, Attributes),
-        NewPrivate = validate_on_load_attribute(OnLoadAttribute, AllDefinitions, Private, Line, E),
+        validate_on_load_attribute(OnLoadAttribute, AllDefinitions, DataBag, Line, E),
 
         DialyzerAttribute = lists:keyfind(dialyzer, 1, Attributes),
         validate_dialyzer_attribute(DialyzerAttribute, AllDefinitions, Line, E),
 
         NifsAttribute = lists:keyfind(nifs, 1, Attributes),
         validate_nifs_attribute(NifsAttribute, AllDefinitions, Line, E),
-
-        Unreachable = elixir_locals:warn_unused_local(Module, AllDefinitions, NewPrivate, E),
-        elixir_locals:ensure_no_undefined_local(Module, AllDefinitions, E),
-        elixir_locals:ensure_no_import_conflict(Module, AllDefinitions, E),
-
-        %% We stop tracking locals here to avoid race conditions in case after_load
-        %% evaluates code in a separate process that may write to locals table.
-        elixir_locals:stop({DataSet, DataBag}),
+        elixir_import:ensure_no_local_conflict(Module, AllDefinitions, E),
         make_readonly(Module),
 
         (not elixir_config:is_bootstrap()) andalso
          'Elixir.Module':'__check_attributes__'(E, DataSet, DataBag),
 
+        AfterVerify = bag_lookup_element(DataBag, {accumulate, after_verify}, 2),
+        [elixir_env:trace({remote_function, [{line, Line}], VerifyMod, VerifyFun, 1}, CallbackE) ||
+         {VerifyMod, VerifyFun} <- AfterVerify],
+
+        %% Ensure there are no errors before we infer types
+        compile_error_if_tainted(DataSet, E),
+
+        {Signatures, Unreachable} =
+          case elixir_config:is_bootstrap() of
+            true -> {#{}, []};
+            false ->
+              UsedPrivate = bag_lookup_element(DataBag, used_private, 2),
+              'Elixir.Module.Types':infer(Module, File, AllDefinitions, Private, UsedPrivate, E, CheckerInfo)
+          end,
+
         RawCompileOpts = bag_lookup_element(DataBag, {accumulate, compile}, 2),
         CompileOpts = validate_compile_opts(RawCompileOpts, AllDefinitions, Unreachable, Line, E),
-        UsesBehaviours = bag_lookup_element(DataBag, {accumulate, behaviour}, 2),
         Impls = bag_lookup_element(DataBag, impls, 2),
 
-        AfterVerify = bag_lookup_element(DataBag, {accumulate, after_verify}, 2),
-        [elixir_env:trace({remote_function, [], VerifyMod, VerifyFun, 1}, CallbackE) ||
-         {VerifyMod, VerifyFun} <- AfterVerify],
-
         ModuleMap = #{
           struct => get_struct(DataSet),
           module => Module,
-          line => Line,
+          anno => Anno,
           file => File,
           relative_file => elixir_utils:relative_to_cwd(File),
           attributes => Attributes,
           definitions => AllDefinitions,
-          unreachable => Unreachable,
           after_verify => AfterVerify,
           compile_opts => CompileOpts,
           deprecated => get_deprecated(DataBag),
           defines_behaviour => defines_behaviour(DataBag),
-          uses_behaviours => UsesBehaviours,
-          impls => Impls
+          impls => Impls,
+          unreachable => Unreachable,
+          signatures => Signatures
         },
 
-        case ets:member(DataSet, {elixir, taint}) of
-          true -> elixir_errors:compile_error(E);
-          false -> ok
-        end,
-
+        compile_error_if_tainted(DataSet, E),
         Binary = elixir_erl:compile(ModuleMap),
         Autoload = proplists:get_value(autoload, CompileOpts, true),
         spawn_parallel_checker(CheckerInfo, Module, ModuleMap),
         {Binary, PersistedAttributes, Autoload}
       end),
 
-    Autoload andalso code:load_binary(Module, beam_location(ModuleAsCharlist), Binary),
+    Autoload andalso code:load_binary(Module, BeamLocation, Binary),
     put_compiler_modules(CompilerModules),
     eval_callbacks(Line, DataBag, after_compile, [CallbackE, Binary], CallbackE),
     elixir_env:trace({on_module, Binary, none}, ModuleE),
     warn_unused_attributes(DataSet, DataBag, PersistedAttributes, E),
     make_module_available(Module, Binary),
-    (CheckerInfo == undefined) andalso
+    (element(2, CheckerInfo) == nil) andalso
       [VerifyMod:VerifyFun(Module) ||
        {VerifyMod, VerifyFun} <- bag_lookup_element(DataBag, {accumulate, after_verify}, 2)],
     {module, Module, Binary, Result}
@@ -220,6 +247,12 @@ compile(Line, Module, ModuleAsCharlist, Block, Vars, Prune, E) ->
     elixir_code_server:call({undefmodule, Ref})
   end.
 
+compile_error_if_tainted(DataSet, E) ->
+  case ets:member(DataSet, {elixir, taint}) of
+    true -> elixir_errors:compile_error(E);
+    false -> ok
+  end.
+
 validate_compile_opts(Opts, Defs, Unreachable, Line, E) ->
   lists:flatmap(fun (Opt) -> validate_compile_opt(Opt, Defs, Unreachable, Line, E) end, Opts).
 
@@ -246,7 +279,7 @@ validate_inlines([Inline | Inlines], Defs, Unreachable, Acc) ->
   case lists:keyfind(Inline, 1, Defs) of
     false ->
       {error, {undefined_function, {compile, inline}, Inline}};
-    {_Def, Type, _Meta, _Clauses} when Type == defmacro; Type == defmacrop ->
+    {_Def, Kind, _Meta, _Clauses} when Kind == defmacro; Kind == defmacrop ->
       {error, {bad_macro, {compile, inline}, Inline}};
     _ ->
       case lists:member(Inline, Unreachable) of
@@ -256,18 +289,17 @@ validate_inlines([Inline | Inlines], Defs, Unreachable, Acc) ->
   end;
 validate_inlines([], _Defs, _Unreachable, Acc) -> {ok, Acc}.
 
-validate_on_load_attribute({on_load, Def}, Defs, Private, Line, E) ->
+validate_on_load_attribute({on_load, Def}, Defs, Bag, Line, E) ->
   case lists:keyfind(Def, 1, Defs) of
     false ->
-      elixir_errors:module_error([{line, Line}], E, ?MODULE, {undefined_function, on_load, Def}),
-      Private;
-    {_Def, Type, _Meta, _Clauses} when Type == defmacro; Type == defmacrop ->
-      elixir_errors:module_error([{line, Line}], E, ?MODULE, {bad_macro, on_load, Def}),
-      Private;
-    _ ->
-      lists:keydelete(Def, 1, Private)
+      elixir_errors:module_error([{line, Line}], E, ?MODULE, {undefined_function, on_load, Def});
+    {_Def, Kind, _Meta, _Clauses} when Kind == defmacro; Kind == defmacrop ->
+      elixir_errors:module_error([{line, Line}], E, ?MODULE, {bad_macro, on_load, Def});
+    {{Name, Arity}, Kind, _Meta, _Clauses} ->
+      elixir_env:trace({local_function, [{line, Line}], Name, Arity}, E),
+      (Kind == defp) andalso ets:insert(Bag, {used_private, Def})
   end;
-validate_on_load_attribute(false, _Defs, Private, _Line, _E) -> Private.
+validate_on_load_attribute(false, _Defs, _Bag, _Line, _E) -> ok.
 
 validate_dialyzer_attribute({dialyzer, Dialyzer}, Defs, Line, E) ->
   [validate_definition({dialyzer, Key}, Fun, Defs, Line, E)
@@ -343,7 +375,6 @@ build(Module, Line, File, E) ->
   %% * {{type, Tuple}, ...}, {{opaque, Tuple}, ...}
   %% * {{callback, Tuple}, ...}, {{macrocallback, Tuple}, ...}
   %% * {{def, Tuple}, ...} (from elixir_def)
-  %% * {{import, Tuple}, ...} (from elixir_locals)
   %% * {{overridable, Tuple}, ...} (from elixir_overridable)
   %%
   DataSet = ets:new(Module, [set, public]),
@@ -359,8 +390,6 @@ build(Module, Line, File, E) ->
   %% * {overridables, ...} (from elixir_overridable)
   %% * {{default, Name}, ...} (from elixir_def)
   %% * {{clauses, Tuple}, ...} (from elixir_def)
-  %% * {reattach, ...} (from elixir_locals)
-  %% * {{local, Tuple}, ...} (from elixir_locals)
   %%
   DataBag = ets:new(Module, [duplicate_bag, public]),
 
@@ -387,6 +416,7 @@ build(Module, Line, File, E) ->
     {optional_callbacks, [], accumulate, []},
 
     % Others
+    {?cache_key, 0},
     {?counter_attr, 0}
   ]),
 
@@ -403,7 +433,6 @@ build(Module, Line, File, E) ->
   %% Setup definition related modules
   Tables = {DataSet, DataBag},
   elixir_def:setup(Tables),
-  elixir_locals:setup(Tables),
   Tuple = {Module, Tables, Line, File, all},
 
   Ref =
@@ -457,8 +486,8 @@ expand_callback(Line, M, F, Args, Acc, Fun) ->
   Meta = [{line, Line}, {required, true}],
 
   {EE, _S, ET} =
-    elixir_dispatch:dispatch_require(Meta, M, F, Args, S, E, fun(AM, AF, AA) ->
-      Fun(AM, AF, AA),
+    elixir_dispatch:dispatch_require(Meta, M, F, Args, S, E, fun(AM, AF) ->
+      Fun(AM, AF, Args),
       {ok, S, E}
     end),
 
@@ -501,7 +530,7 @@ warn_unused_attributes(DataSet, DataBag, PersistedAttrs, E) ->
 get_struct(Set) ->
   case ets:lookup(Set, {elixir, struct}) of
     [] -> nil;
-    [{_, Struct}] -> Struct
+    [{_, Fields}] -> Fields
   end.
 
 get_deprecated(Bag) ->
@@ -526,11 +555,11 @@ beam_location(ModuleAsCharlist) ->
 
 checker_info() ->
   case get(elixir_checker_info) of
-    undefined -> undefined;
+    undefined -> {self(), nil};
     _ -> 'Elixir.Module.ParallelChecker':get()
   end.
 
-spawn_parallel_checker(undefined, _Module, _ModuleMap) ->
+spawn_parallel_checker({_, nil}, _Module, _ModuleMap) ->
   ok;
 spawn_parallel_checker(CheckerInfo, Module, ModuleMap) ->
   Log =
