@@ -1,6 +1,6 @@
 -module(elixir_quote).
 -export([escape/3, linify/3, linify_with_context_counter/3, build/7, quote/2, has_unquotes/1, fun_to_quoted/1]).
--export([dot/5, tail_list/3, list/2, validate_runtime/2]). %% Quote callbacks
+-export([dot/5, tail_list/3, list/2, validate_runtime/2, shallow_validate_ast/1]). %% Quote callbacks
 
 -include("elixir.hrl").
 -define(defs(Kind), Kind == def; Kind == defp; Kind == defmacro; Kind == defmacrop; Kind == '@').
@@ -15,7 +15,8 @@
   aliases_hygiene=nil,
   imports_hygiene=nil,
   unquote=true,
-  generated=false
+  generated=false,
+  shallow_validate=false
 }).
 
 %% fun_to_quoted
@@ -140,7 +141,7 @@ escape(Expr, Op, Unquote) ->
   }).
 
 do_escape({Left, Meta, Right}, #elixir_quote{op=prune_metadata} = Q) when is_list(Meta) ->
-  TM = [{K, V} || {K, V} <- Meta, (K == no_parens) orelse (K == line)],
+  TM = [{K, V} || {K, V} <- Meta, (K == no_parens) orelse (K == line) orelse (K == delimiter)],
   TL = do_quote(Left, Q),
   TR = do_quote(Right, Q),
   {'{}', [], [TL, TM, TR]};
@@ -159,10 +160,19 @@ do_escape(BitString, _) when is_bitstring(BitString) ->
   end;
 
 do_escape(Map, Q) when is_map(Map) ->
-  TT = do_quote(lists:sort(maps:to_list(Map)), Q),
+  TT =
+    [if
+      is_reference(V) ->
+        argument_error(<<('Elixir.Kernel':inspect(Map, []))/binary, " contains a reference (",
+                         ('Elixir.Kernel':inspect(V, []))/binary, ") and therefore it cannot be escaped ",
+                         "(it must be defined within a function instead). ", (bad_escape_hint())/binary>>);
+      true ->
+        {do_quote(K, Q), do_quote(V, Q)}
+     end || {K, V} <- lists:sort(maps:to_list(Map))],
   {'%{}', [], TT};
 
-do_escape([], _) -> [];
+do_escape([], _) ->
+  [];
 
 do_escape([H | T], #elixir_quote{unquote=false} = Q) ->
   do_quote_simple_list(T, do_quote(H, Q), Q);
@@ -179,8 +189,7 @@ do_escape([H | T], Q) ->
       update_last(TL, fun(X) -> {'|', [], [X, TR]} end)
   end;
 
-do_escape(Other, _)
-    when is_number(Other); is_pid(Other); is_atom(Other) ->
+do_escape(Other, _) when is_number(Other); is_atom(Other); is_pid(Other) ->
   Other;
 
 do_escape(Fun, _) when is_function(Fun) ->
@@ -195,8 +204,11 @@ do_escape(Other, _) ->
 
 bad_escape(Arg) ->
   argument_error(<<"cannot escape ", ('Elixir.Kernel':inspect(Arg, []))/binary, ". ",
-                   "The supported values are: lists, tuples, maps, atoms, numbers, bitstrings, ",
-                   "PIDs and remote functions in the format &Mod.fun/arity">>).
+                   (bad_escape_hint())/binary>>).
+
+bad_escape_hint() ->
+  <<"The supported values are: lists, tuples, maps, atoms, numbers, bitstrings, ",
+    "PIDs and remote functions in the format &Mod.fun/arity">>.
 
 %% Quote entry points
 
@@ -216,7 +228,8 @@ build(Meta, Line, File, Context, Unquote, Generated, E) ->
     file=VFile,
     unquote=Unquote,
     context=VContext,
-    generated=Generated
+    generated=Generated,
+    shallow_validate=true
   },
 
   {Q, VContext, Acc3}.
@@ -255,6 +268,27 @@ is_valid(context, Context) -> is_atom(Context) andalso (Context /= nil);
 is_valid(generated, Generated) -> is_boolean(Generated);
 is_valid(unquote, Unquote) -> is_boolean(Unquote).
 
+shallow_validate_ast(Expr) ->
+  case shallow_valid_ast(Expr) of
+    true -> Expr;
+    false -> argument_error(
+      <<"tried to unquote invalid AST: ", ('Elixir.Kernel':inspect(Expr))/binary,
+        "\nDid you forget to escape term using Macro.escape/1?">>)
+  end.
+
+shallow_valid_ast(Expr) when is_list(Expr) -> valid_ast_list(Expr);
+shallow_valid_ast(Expr) -> valid_ast_elem(Expr).
+
+valid_ast_list([]) -> true;
+valid_ast_list([Head | Tail]) -> valid_ast_elem(Head) andalso valid_ast_list(Tail);
+valid_ast_list(_Improper) -> false.
+
+valid_ast_elem(Expr) when is_list(Expr); is_atom(Expr); is_binary(Expr); is_number(Expr); is_pid(Expr); is_function(Expr) -> true;
+valid_ast_elem({Left, Right}) -> valid_ast_elem(Left) andalso valid_ast_elem(Right);
+valid_ast_elem({Atom, Meta, Args}) when is_atom(Atom), is_list(Meta), is_atom(Args) orelse is_list(Args) -> true;
+valid_ast_elem({Call, Meta, Args}) when is_list(Meta), is_list(Args) -> shallow_valid_ast(Call);
+valid_ast_elem(_Term) -> false.
+
 quote({unquote_splicing, _, [_]}, #elixir_quote{unquote=true}) ->
   argument_error(<<"unquote_splicing only works inside arguments and block contexts, "
     "wrap it in parens if you want it to work with one-liners">>);
@@ -284,8 +318,12 @@ do_quote({quote, Meta, [Opts, Arg]}, Q) when is_list(Meta) ->
 
   {'{}', [], [quote, meta(NewMeta, Q), [TOpts, TArg]]};
 
-do_quote({unquote, Meta, [Expr]}, #elixir_quote{unquote=true}) when is_list(Meta) ->
-  Expr;
+%
+do_quote({unquote, Meta, [Expr]}, #elixir_quote{unquote=true, shallow_validate=Validate}) when is_list(Meta) ->
+  case Validate of
+    true -> {{'.', Meta, [?MODULE, shallow_validate_ast]}, Meta, [Expr]};
+    false -> Expr
+  end;
 
 %% Aliases
 
@@ -380,6 +418,7 @@ import_meta(Meta, Name, Arity, Q, E) ->
   case (keyfind(imports, Meta) == false) andalso
       elixir_dispatch:find_imports(Meta, Name, E) of
     [_ | _] = Imports ->
+      trace_import_quoted(Imports, Meta, Name, E),
       keystore(imports, keystore(context, Meta, Q#elixir_quote.context), Imports);
 
     _ ->
@@ -389,6 +428,20 @@ import_meta(Meta, Name, Arity, Q, E) ->
       end
   end.
 
+trace_import_quoted([{Arity, Mod} | Imports], Meta, Name, E) ->
+  {Rest, Arities} = collect_trace_import_quoted(Imports, Mod, [], [Arity]),
+  elixir_env:trace({imported_quoted, Meta, Mod, Name, Arities}, E),
+  trace_import_quoted(Rest, Meta, Name, E);
+trace_import_quoted([], _Meta, _Name, _E) ->
+  ok.
+
+collect_trace_import_quoted([{Arity, Mod} | Imports], Mod, Acc, Arities) ->
+  collect_trace_import_quoted(Imports, Mod, Acc, [Arity | Arities]);
+collect_trace_import_quoted([Import | Imports], Mod, Acc, Arities) ->
+  collect_trace_import_quoted(Imports, Mod, [Import | Acc], Arities);
+collect_trace_import_quoted([], _Mod, Acc, Arities) ->
+  {lists:reverse(Acc), lists:reverse(Arities)}.
+
 %% do_quote_*
 
 do_quote_call(Left, Meta, Expr, Args, Q) ->
@@ -490,11 +543,12 @@ tail_list(Left, Right, Tail) when is_list(Left) ->
   [H | T] = lists:reverse(Tail ++ Left),
   lists:reverse([{'|', [], [H, Right]} | T]).
 
-validate_list(List) when is_list(List) ->
-  ok;
-validate_list(List) when not is_list(List) ->
-  argument_error(<<"expected a list with quoted expressions in unquote_splicing/1, got: ",
-                   ('Elixir.Kernel':inspect(List))/binary>>).
+validate_list(List) ->
+  case valid_ast_list(List) of
+    true -> ok;
+    false -> argument_error(<<"expected a list with quoted expressions in unquote_splicing/1, got: ",
+                   ('Elixir.Kernel':inspect(List))/binary>>)
+  end.
 
 argument_error(Message) ->
   error('Elixir.ArgumentError':exception([{message, Message}])).
