@@ -46,6 +46,12 @@ defmodule ExceptionTest do
                {:io, :put_chars, [self(), <<222>>],
                 [error_info: %{module: __MODULE__, function: :dummy_error_extras}]}
              ])
+
+    assert %ErlangError{original: {:failed_load_cacerts, :enoent}, reason: ": this is chardata"} =
+             Exception.normalize(:error, {:failed_load_cacerts, :enoent}, [
+               {:pubkey_os_cacerts, :get, 0,
+                [error_info: %{module: __MODULE__, function: :dummy_error_chardata}]}
+             ])
   end
 
   test "format/2 without stacktrace" do
@@ -430,8 +436,8 @@ defmodule ExceptionTest do
         end
       )
 
-      :code.delete(OperatorPrecedence)
       :code.purge(OperatorPrecedence)
+      :code.delete(OperatorPrecedence)
 
       assert blame_message(OperatorPrecedence, & &1.test!(1, 2)) =~ """
              no function clause matching in ExceptionTest.OperatorPrecedence.test!/2
@@ -469,8 +475,8 @@ defmodule ExceptionTest do
         end
       )
 
-      :code.delete(Req)
       :code.purge(Req)
+      :code.delete(Req)
 
       assert blame_message(Req, & &1.get!(url: "https://elixir-lang.org")) =~ """
              no function clause matching in ExceptionTest.Req.get!/1
@@ -613,7 +619,8 @@ defmodule ExceptionTest do
              """
 
       assert blame_message(ENUM, & &1.not_a_function(&1, 1)) ==
-               "function ENUM.not_a_function/2 is undefined (module ENUM is not available)"
+               "function ENUM.not_a_function/2 is undefined (module ENUM is not available). " <>
+                 "Make sure the module name is correct and has been specified in full (or that an alias has been defined)"
 
       assert blame_message(One, & &1.foo()) == """
              function One.foo/0 is undefined (module One is not available). Did you mean:
@@ -623,8 +630,8 @@ defmodule ExceptionTest do
              """
 
       for module <- modules do
-        :code.delete(module)
         :code.purge(module)
+        :code.delete(module)
       end
     end
 
@@ -839,8 +846,8 @@ defmodule ExceptionTest do
         end
       )
 
-      :code.delete(Blaming)
       :code.purge(Blaming)
+      :code.delete(Blaming)
 
       {:ok, :def, clauses} = Exception.blame_mfa(Blaming, :with_elem, [1, 2])
 
@@ -935,7 +942,9 @@ defmodule ExceptionTest do
              |> message == "function Kernel.bar/1 is undefined or private"
 
       assert %UndefinedFunctionError{module: Foo, function: :bar, arity: 1}
-             |> message == "function Foo.bar/1 is undefined (module Foo is not available)"
+             |> message ==
+               "function Foo.bar/1 is undefined (module Foo is not available). " <>
+                 "Make sure the module name is correct and has been specified in full (or that an alias has been defined)"
 
       assert %UndefinedFunctionError{module: nil, function: :bar, arity: 3}
              |> message == "function nil.bar/3 is undefined"
@@ -954,6 +963,21 @@ defmodule ExceptionTest do
     test "ErlangError" do
       assert %ErlangError{original: :sample} |> message == "Erlang error: :sample"
     end
+
+    test "MissingApplicationsError" do
+      assert %MissingApplicationsError{
+               apps: [{:logger, "~> 1.18"}, {:ex_unit, Version.parse_requirement!(">= 0.0.0")}],
+               description: "applications are required"
+             }
+             |> message == """
+             applications are required
+
+             To address this, include these applications as your dependencies:
+
+               {:logger, "~> 1.18"}
+               {:ex_unit, ">= 0.0.0"}\
+             """
+    end
   end
 
   describe "error_info" do
@@ -1008,4 +1032,8 @@ defmodule ExceptionTest do
   end
 
   def dummy_error_extras(_exception, _stacktrace), do: %{general: "foo"}
+
+  def dummy_error_chardata(_exception, _stacktrace) do
+    %{general: ~c"this is " ++ [~c"chardata"], reason: ~c"this " ++ [~c"too"]}
+  end
 end
