@@ -52,7 +52,6 @@ defmodule Kernel.LexicalTrackerTest do
   end
 
   test "can add aliases", config do
-    D.add_alias(config[:pid], String, 1, true)
     D.alias_dispatch(config[:pid], String)
     assert D.references(config[:pid]) == {[], [], [], []}
   end
@@ -93,18 +92,19 @@ defmodule Kernel.LexicalTrackerTest do
   end
 
   test "unused aliases", config do
-    D.add_alias(config[:pid], String, 1, true)
-    assert D.collect_unused_aliases(config[:pid]) == [{String, 1}]
+    D.warn_alias(config[:pid], [], String, String)
+    assert D.collect_unused_aliases(config[:pid]) == [{String, []}]
   end
 
   test "used aliases are not unused", config do
-    D.add_alias(config[:pid], String, 1, true)
+    D.warn_alias(config[:pid], [], String, String)
     D.alias_dispatch(config[:pid], String)
     assert D.collect_unused_aliases(config[:pid]) == []
   end
 
-  test "aliases with no warn are not unused", config do
-    D.add_alias(config[:pid], String, 1, false)
+  test "used aliases are not unused in reverse order", config do
+    D.alias_dispatch(config[:pid], String)
+    D.warn_alias(config[:pid], [], String, String)
     assert D.collect_unused_aliases(config[:pid]) == []
   end
 
@@ -411,7 +411,7 @@ defmodule Kernel.LexicalTrackerTest do
       assert URI in runtime
     end
 
-    test "Macro.struct! adds an export dependency" do
+    test "Macro.struct_info! adds an export dependency" do
       {{compile, exports, runtime, _}, _binding} =
         Code.eval_string("""
         defmodule Kernel.LexicalTrackerTest.MacroStruct do
@@ -420,7 +420,7 @@ defmodule Kernel.LexicalTrackerTest do
           # mechanism to make this expansion become a compile-time one.
           # However, in some cases, such as typespecs, we don't necessarily
           # want the compile-time dependency to happen.
-          Macro.struct!(:"Elixir.URI", __ENV__)
+          Macro.struct_info!(:"Elixir.URI", __ENV__)
           Kernel.LexicalTracker.references(__ENV__.lexical_tracker)
         end |> elem(3)
         """)
@@ -517,5 +517,55 @@ defmodule Kernel.LexicalTrackerTest do
       refute URI in exports
       refute URI in runtime
     end
+
+    test "imported functions from quote adds dependencies" do
+      {{compile, exports, runtime, _}, _binding} =
+        Code.eval_string("""
+        defmodule Kernel.LexicalTrackerTest.QuotedFun do
+          import URI
+
+          defmacro parse_root() do
+            quote do
+              parse("/")
+            end
+          end
+        end
+
+        defmodule Kernel.LexicalTrackerTest.UsingQuotedFun do
+          require Kernel.LexicalTrackerTest.QuotedFun, as: QF
+          QF.parse_root()
+          Kernel.LexicalTracker.references(__ENV__.lexical_tracker)
+        end |> elem(3)
+        """)
+
+      assert URI in compile
+      refute URI in exports
+      refute URI in runtime
+    end
+
+    test "imported macro from quote adds dependencies" do
+      {{compile, exports, runtime, _}, _binding} =
+        Code.eval_string("""
+        defmodule Kernel.LexicalTrackerTest.QuotedMacro do
+          import Config
+
+          defmacro config_env() do
+            quote do
+              config_env()
+            end
+          end
+        end
+
+        defmodule Kernel.LexicalTrackerTest.UsingQuotedMacro do
+          require Kernel.LexicalTrackerTest.QuotedMacro, as: QM
+          def fun(), do: QM.config_env()
+          Kernel.LexicalTracker.references(__ENV__.lexical_tracker)
+        end |> elem(3)
+        """)
+
+      assert Config in compile
+      refute Config in exports
+      refute Config in runtime
+    end
   end
 end
