@@ -84,13 +84,6 @@ defmodule ListTest do
     assert List.last([1, 2, 3]) == 3
   end
 
-  test "zip/1" do
-    assert List.zip([[1, 4], [2, 5], [3, 6]]) == [{1, 2, 3}, {4, 5, 6}]
-    assert List.zip([[1, 4], [2, 5, 0], [3, 6]]) == [{1, 2, 3}, {4, 5, 6}]
-    assert List.zip([[1], [2, 5], [3, 6]]) == [{1, 2, 3}]
-    assert List.zip([[1, 4], [2, 5], []]) == []
-  end
-
   test "keyfind/4" do
     assert List.keyfind([a: 1, b: 2], :a, 0) == {:a, 1}
     assert List.keyfind([a: 1, b: 2], 2, 1) == {:b, 2}
@@ -283,11 +276,36 @@ defmodule ListTest do
       assert_raise FunctionClauseError, message, fn ->
         List.starts_with?([1 | 2], [1 | 2])
       end
+    end
+  end
 
-      message = "no function clause matching in List.starts_with?/2"
+  describe "ends_with?/2" do
+    test "list and prefix are equal" do
+      assert List.ends_with?([], [])
+      assert List.ends_with?([1], [1])
+      assert List.ends_with?([1, 2, 3], [1, 2, 3])
+    end
 
-      assert_raise FunctionClauseError, message, fn ->
-        List.starts_with?([1, 2], 1)
+    test "proper lists" do
+      refute List.ends_with?([2], [1, 2])
+      assert List.ends_with?([1, 2, 3], [2, 3])
+      refute List.ends_with?([2, 3, 4], [1, 2, 3, 4])
+    end
+
+    test "list is empty" do
+      refute List.ends_with?([], [1])
+      refute List.ends_with?([], [1, 2])
+    end
+
+    test "prefix is empty" do
+      assert List.ends_with?([1], [])
+      assert List.ends_with?([1, 2], [])
+      assert List.ends_with?([1, 2, 3], [])
+    end
+
+    test "only accepts proper lists" do
+      assert_raise ArgumentError, ~r/not a list/, fn ->
+        List.ends_with?([1 | 2], [1 | 2])
       end
     end
   end
