@@ -352,6 +352,11 @@ defmodule Mix do
     * `MIX_INSTALL_DIR` *(since v1.12.0)* - specifies directory where `Mix.install/2` keeps
        install cache
 
+    * `MIX_OS_CONCURRENCY_LOCK` - when set to `0` or `false`, disables mix compilation locking.
+      While not recommended, this may be necessary in cases where hard links or TCP sockets are
+      not available. When opting for this behaviour, make sure to not start concurrent compilations
+      of the same project.
+
     * `MIX_PATH` - appends extra code paths
 
     * `MIX_PROFILE` - a list of comma-separated Mix tasks to profile the time spent on
@@ -375,13 +380,38 @@ defmodule Mix do
 
       $ MIX_DEBUG=1 mix compile
 
-  In addition, Mix also uses the following environment variables defined by other libraries
+  ## SSL certificates
+
+  Mix and the Hex package manager use the default operating system certificates
+  when downloading resources. In certain situations, such as when running behind
+  proxies, you want to replace those certificates.
+
+  If you simply want to change the certificates used by Mix and Hex, you may set
+  the `HEX_CACERTS_PATH` environment variable, pointing to a CA certificate file.
+  See [`mix hex.config`](https://hexdocs.pm/hex/Mix.Tasks.Hex.Config.html#module-config-keys).
+
+  From Erlang/OTP 27.2, it is also possible to change the certificates for your
+  project as a whole. To do so, you might add the following to your `config/config.exs`:
+
+      config :public_key, :cacerts_path, "/path/to/certs.pem"
+
+  You can also do so by setting the `ERL_AFLAGS` and `ERL_ZFLAGS` environment variables:
+
+  ```bash
+  ERL_AFLAGS="-public_key cacerts_path '\"/path/to/certs.pem\"'"
+  ERL_ZFLAGS="-public_key cacerts_path '\"/path/to/certs.pem\"'"
+  ```
+
+  You can verify if the configuration has been properly set by calling the following
+  inside `iex -S mix`:
+
+      Application.load(:public_key)
+      Application.get_env(:public_key, :cacerts_path)
+
+  And by loading the certificates:
+
+      :public_key.cacerts_get()
 
-    * `HEX_CACERTS_PATH` - use specified CA certificate file instead of default
-      system CA certificates. This configures how HTTPS calls are made via
-      [Erlang `ssl` module](https://www.erlang.org/doc/man/ssl#type-client_cafile)
-      to fetch remote archives and packages. For more details, see
-      [`mix hex.config`](https://hexdocs.pm/hex/Mix.Tasks.Hex.Config.html#module-config-keys).
   """
 
   @mix_install_project __MODULE__.InstallProject
@@ -402,7 +432,7 @@ defmodule Mix do
   def start(_type, []) do
     Mix.Local.append_archives()
     Mix.Local.append_paths()
-    children = [Mix.State, Mix.TasksServer, Mix.ProjectStack]
+    children = [Mix.Sync.PubSub, Mix.State, Mix.TasksServer, Mix.ProjectStack]
     opts = [strategy: :one_for_one, name: Mix.Supervisor, max_restarts: 0]
     Supervisor.start_link(children, opts)
   end
@@ -640,7 +670,7 @@ defmodule Mix do
             )
 
           %{} ->
-            unless app in optional do
+            if app not in optional do
               Mix.raise(
                 "The application \"#{app}\" could not be found. This may happen if your " <>
                   "Operating System broke Erlang into multiple packages and may be fixed " <>
@@ -998,7 +1028,7 @@ defmodule Mix do
         _ -> Mix.raise("unknown dependency #{inspect(app_name)} given to #{inspect(key)}")
       end
 
-    unless app_dir do
+    if !app_dir do
       Mix.raise("#{inspect(app_name)} given to #{inspect(key)} must be a path dependency")
     end
 
