@@ -75,6 +75,14 @@ defmodule Mix.Project do
   the `:erlc_paths` configuration is used by `mix compile.erlang`, `mix compile.yecc`,
   and other tasks.
 
+  > #### Keep `project/0` fast {: .warning}
+  >
+  > `project/0` is called whenever your `mix.exs` is loaded, so heavy
+  > computation should be avoided. If a task requires a potentially complex
+  > configuration value, it should allow its configuration to be set to an
+  > anonymous function or similar, so that it can be invoked only when
+  > needed by the task itself.
+
   ## CLI configuration
 
   Mix is most often invoked from the command line. For this purpose, you may define
@@ -163,7 +171,12 @@ defmodule Mix.Project do
   # Only the top of the stack can be accessed.
   @doc false
   def push(module, file \\ nil, app \\ nil) when is_atom(module) do
-    file = file || (module && List.to_string(module.__info__(:compile)[:source]))
+    file =
+      cond do
+        file != nil -> file
+        source = module && module.module_info(:compile)[:source] -> List.to_string(source)
+        true -> "nofile"
+      end
 
     case Mix.ProjectStack.push(module, push_config(module, app), file) do
       :ok ->
@@ -188,13 +201,13 @@ defmodule Mix.Project do
 
       task = String.to_atom(task || config[:default_task] || "run")
 
-      unless System.get_env("MIX_ENV") do
+      if !System.get_env("MIX_ENV") do
         if env = config[:preferred_envs][task] || @preferred_envs[task] || config[:default_env] do
           Mix.env(env)
         end
       end
 
-      unless System.get_env("MIX_TARGET") do
+      if !System.get_env("MIX_TARGET") do
         if target = config[:preferred_targets][task] || config[:default_target] do
           Mix.target(target)
         end
@@ -882,6 +895,54 @@ defmodule Mix.Project do
     end
   end
 
+  @doc """
+  Acquires a lock on the project build path and runs the given function.
+
+  When another process (across all OS processes) is holding the lock,
+  a message is printed and this call blocks until the lock is acquired.
+  This function can also be called if this process already has the
+  lock. In such case the function is executed immediately.
+
+  This lock is primarily useful for compiler tasks that alter the build
+  artifacts to avoid conflicts with a concurrent compilation.
+  """
+  @spec with_build_lock(keyword, (-> term())) :: term()
+  def with_build_lock(config \\ config(), fun) do
+    # To avoid duplicated compilation, we wrap compilation tasks, such
+    # as compile.all, deps.compile, compile.elixir, compile.erlang in
+    # a lock. Note that compile.all covers compile.elixir, but the
+    # latter can still be invoked directly, so we put the lock over
+    # each individual task.
+
+    build_path = build_path(config)
+
+    on_taken = fn os_pid ->
+      Mix.shell().error([
+        IO.ANSI.reset(),
+        "Waiting for lock on the build directory (held by process #{os_pid})"
+      ])
+    end
+
+    Mix.Sync.Lock.with_lock(build_path, fun, on_taken: on_taken)
+  end
+
+  @doc false
+  def with_deps_lock(config \\ config(), fun) do
+    # We wrap operations on the deps directory and on mix.lock to
+    # avoid write conflicts.
+
+    deps_path = deps_path(config)
+
+    on_taken = fn os_pid ->
+      Mix.shell().error([
+        IO.ANSI.reset(),
+        "Waiting for lock on the deps directory (held by process #{os_pid})"
+      ])
+    end
+
+    Mix.Sync.Lock.with_lock(deps_path, fun, on_taken: on_taken)
+  end
+
   # Loads mix.exs in the current directory or loads the project from the
   # mixfile cache and pushes the project onto the project stack.
   defp load_project(app, post_config) do
