@@ -92,6 +92,13 @@ defmodule Mix.Tasks.Compile.App do
     * `--compile-path` - where to find `.beam` files and write the
       resulting `.app` file, defaults to `Mix.Project.compile_path/0`
 
+  ## Configuration
+
+    * `:reliable_dir_mtime` - this task relies on the operating system
+      changing the mtime on a directory whenever a file is added or removed.
+      You can set this option to false if your system does not provide
+      reliable mtimes. Defaults to false on Windows.
+
   ## Phases
 
   Applications provide a start phases mechanism which will be called,
@@ -130,50 +137,65 @@ defmodule Mix.Tasks.Compile.App do
   @impl true
   def run(args) do
     {opts, _, _} = OptionParser.parse(args, switches: [force: :boolean, compile_path: :string])
-
     project = Mix.Project.get!()
     config = Mix.Project.config()
 
     app = Keyword.get(config, :app)
     version = Keyword.get(config, :version)
+    validate_app!(app)
+    validate_version!(version)
 
-    validate_app(app)
-    validate_version(version)
+    compile_path = Keyword.get_lazy(opts, :compile_path, &Mix.Project.compile_path/0)
+    target = Path.join(compile_path, "#{app}.app")
 
-    path = Keyword.get_lazy(opts, :compile_path, &Mix.Project.compile_path/0)
-    modules = modules_from(path) |> Enum.sort()
-
-    target = Path.join(path, "#{app}.app")
-
-    # We mostly depend on the project_file through the def application function,
-    # but it doesn't hurt to also include config_mtime.
+    # If configurations changed, we may have changed compile_env.
+    # If compile_path changed, we may have added or removed files.
+    # If the project changed, we may have changed other properties.
     new_mtime =
-      max(Mix.Project.config_mtime(), Mix.Utils.last_modified(Mix.Project.project_file()))
+      Mix.Project.config_mtime()
+      |> max(Mix.Utils.last_modified(Mix.Project.project_file()))
+      |> max(Mix.Utils.last_modified(compile_path))
 
     current_properties = current_app_properties(target)
-    compile_env = load_compile_env(current_properties)
-    old_mtime = Keyword.get(current_properties, :config_mtime, 0)
 
-    if opts[:force] || new_mtime > old_mtime ||
-         app_changed?(current_properties, modules, compile_env) do
+    {changed?, modules} =
+      cond do
+        opts[:force] || new_mtime > Mix.Utils.last_modified(target) ->
+          {true, nil}
+
+        Keyword.get(config, :reliable_dir_mtime, fn -> not match?({:win32, _}, :os.type()) end) ->
+          {false, nil}
+
+        true ->
+          modules = modules_from(compile_path)
+          {modules != Keyword.get(current_properties, :modules, []), modules}
+      end
+
+    if changed? do
       properties =
         [
           description: to_charlist(config[:description] || app),
-          modules: modules,
           registered: [],
           vsn: to_charlist(version)
         ]
         |> merge_project_application(project)
-        |> validate_properties!()
         |> handle_extra_applications(config)
-        |> add_compile_env(compile_env)
+        |> add_compile_env(current_properties)
+        |> add_modules(modules, compile_path)
 
-      properties = [config_mtime: new_mtime] ++ properties
       contents = :io_lib.format("~p.~n", [{:application, app, properties}])
       :application.load({:application, app, properties})
 
       Mix.Project.ensure_structure()
       File.write!(target, IO.chardata_to_string(contents))
+      File.touch!(target, new_mtime)
+
+      # If we just created the .app file, it will have touched
+      # the directory mtime, so we need to reset it.
+      if current_properties == [] do
+        File.touch!(compile_path, new_mtime)
+      end
+
       Mix.shell().info("Generated #{app} app")
       {:ok, []}
     else
@@ -189,47 +211,37 @@ defmodule Mix.Tasks.Compile.App do
     end
   end
 
-  defp load_compile_env(current_properties) do
-    case Mix.ProjectStack.compile_env(nil) do
-      nil -> Keyword.get(current_properties, :compile_env, [])
-      list -> list
-    end
-  end
+  defp validate_app!(app) when is_atom(app), do: :ok
 
-  defp app_changed?(properties, mods, compile_env) do
-    Keyword.get(properties, :modules, []) != mods or
-      Keyword.get(properties, :compile_env, []) != compile_env
-  end
-
-  defp validate_app(app) when is_atom(app), do: :ok
-
-  defp validate_app(app) do
-    ensure_present(:app, app)
+  defp validate_app!(app) do
+    ensure_present!(:app, app)
     Mix.raise("Expected :app to be an atom, got: #{inspect(app)}")
   end
 
-  defp validate_version(version) do
-    ensure_present(:version, version)
+  defp validate_version!(version) do
+    ensure_present!(:version, version)
 
-    unless is_binary(version) and match?({:ok, _}, Version.parse(version)) do
+    if not (is_binary(version) and match?({:ok, _}, Version.parse(version))) do
       Mix.raise(
         "Expected :version to be a valid Version, got: #{inspect(version)} (see the Version module for more information)"
       )
     end
   end
 
-  defp ensure_present(name, nil) do
+  defp ensure_present!(name, nil) do
     Mix.raise("Please ensure mix.exs file has the #{inspect(name)} in the project definition")
   end
 
-  defp ensure_present(_name, _val), do: :ok
+  defp ensure_present!(_name, _val), do: :ok
 
   defp modules_from(path) do
     case File.ls(path) do
       {:ok, entries} ->
-        for entry <- entries,
-            String.ends_with?(entry, ".beam"),
-            do: entry |> binary_part(0, byte_size(entry) - 5) |> String.to_atom()
+        Enum.sort(
+          for entry <- entries,
+              String.ends_with?(entry, ".beam"),
+              do: entry |> binary_part(0, byte_size(entry) - 5) |> String.to_atom()
+        )
 
       {:error, _} ->
         []
@@ -240,13 +252,13 @@ defmodule Mix.Tasks.Compile.App do
     if function_exported?(project, :application, 0) do
       project_application = project.application()
 
-      unless Keyword.keyword?(project_application) do
+      if not Keyword.keyword?(project_application) do
         Mix.raise(
           "Application configuration returned from application/0 should be a keyword list"
         )
       end
 
-      Keyword.merge(best_guess, project_application)
+      Keyword.merge(best_guess, validate_properties!(project_application))
     else
       best_guess
     end
@@ -255,7 +267,7 @@ defmodule Mix.Tasks.Compile.App do
   defp validate_properties!(properties) do
     Enum.each(properties, fn
       {:description, value} ->
-        unless is_list(value) do
+        if not is_list(value) do
           Mix.raise(
             "Application description (:description) is not a character list, got: " <>
               inspect(value)
@@ -263,17 +275,17 @@ defmodule Mix.Tasks.Compile.App do
         end
 
       {:id, value} ->
-        unless is_list(value) do
+        if not is_list(value) do
           Mix.raise("Application ID (:id) is not a character list, got: " <> inspect(value))
         end
 
       {:vsn, value} ->
-        unless is_list(value) do
+        if not is_list(value) do
           Mix.raise("Application vsn (:vsn) is not a character list, got: " <> inspect(value))
         end
 
       {:maxT, value} ->
-        unless value == :infinity or is_integer(value) do
+        if not (value == :infinity or is_integer(value)) do
           Mix.raise(
             "Application maximum time (:maxT) is not an integer or :infinity, got: " <>
               inspect(value)
@@ -281,14 +293,14 @@ defmodule Mix.Tasks.Compile.App do
         end
 
       {:modules, value} ->
-        unless is_list(value) and Enum.all?(value, &is_atom(&1)) do
+        if not (is_list(value) and Enum.all?(value, &is_atom(&1))) do
           Mix.raise(
             "Application modules (:modules) should be a list of atoms, got: " <> inspect(value)
           )
         end
 
       {:registered, value} ->
-        unless is_list(value) and Enum.all?(value, &is_atom(&1)) do
+        if not (is_list(value) and Enum.all?(value, &is_atom(&1))) do
           Mix.raise(
             "Application registered processes (:registered) should be a list of atoms, got: " <>
               inspect(value)
@@ -296,7 +308,7 @@ defmodule Mix.Tasks.Compile.App do
         end
 
       {:included_applications, value} ->
-        unless is_list(value) and Enum.all?(value, &is_atom(&1)) do
+        if not (is_list(value) and Enum.all?(value, &is_atom(&1))) do
           Mix.raise(
             "Application included applications (:included_applications) should be a list of atoms, got: " <>
               inspect(value)
@@ -304,7 +316,7 @@ defmodule Mix.Tasks.Compile.App do
         end
 
       {:extra_applications, value} ->
-        unless is_list(value) and Enum.all?(value, &typed_app?(&1)) do
+        if not (is_list(value) and Enum.all?(value, &typed_app?(&1))) do
           Mix.raise(
             "Application extra applications (:extra_applications) should be a list of atoms or " <>
               "{app, :required | :optional} pairs, got: " <> inspect(value)
@@ -312,7 +324,7 @@ defmodule Mix.Tasks.Compile.App do
         end
 
       {:applications, value} ->
-        unless is_list(value) and Enum.all?(value, &typed_app?(&1)) do
+        if not (is_list(value) and Enum.all?(value, &typed_app?(&1))) do
           Mix.raise(
             "Application applications (:applications) should be a list of atoms or " <>
               "{app, :required | :optional} pairs, got: " <> inspect(value)
@@ -320,14 +332,14 @@ defmodule Mix.Tasks.Compile.App do
         end
 
       {:env, value} ->
-        unless Keyword.keyword?(value) do
+        if not Keyword.keyword?(value) do
           Mix.raise(
             "Application environment (:env) should be a keyword list, got: " <> inspect(value)
           )
         end
 
       {:start_phases, value} ->
-        unless Keyword.keyword?(value) do
+        if not Keyword.keyword?(value) do
           Mix.raise(
             "Application start phases (:start_phases) should be a keyword list, got: " <>
               inspect(value)
@@ -357,8 +369,21 @@ defmodule Mix.Tasks.Compile.App do
   defp typed_app?({app, type}) when is_atom(app) and type in [:required, :optional], do: true
   defp typed_app?(_), do: false
 
-  defp add_compile_env(properties, []), do: properties
-  defp add_compile_env(properties, compile_env), do: [compile_env: compile_env] ++ properties
+  defp add_compile_env(properties, current_properties) do
+    # If someone calls compile.elixir and then compile.app across two
+    # separate OS calls, then the compile_env won't be properly reflected.
+    # This is ok because compile_env is not used for correctness. It is
+    # simply to catch possible errors early.
+    case Mix.ProjectStack.compile_env(nil) do
+      nil -> Keyword.take(current_properties, [:compile_env]) ++ properties
+      [] -> properties
+      compile_env -> Keyword.put(properties, :compile_env, compile_env)
+    end
+  end
+
+  defp add_modules(properties, modules, compile_path) do
+    Keyword.put_new_lazy(properties, :modules, fn -> modules || modules_from(compile_path) end)
+  end
 
   defp handle_extra_applications(properties, config) do
     {extra, properties} = Keyword.pop(properties, :extra_applications, [])
