@@ -48,20 +48,25 @@ defmodule Mix.Tasks.Deps.Compile do
 
   @impl true
   def run(args) do
-    unless "--no-archives-check" in args do
+    if "--no-archives-check" not in args do
       Mix.Task.run("archive.check", args)
     end
 
     Mix.Project.get!()
-    deps = Mix.Dep.load_and_cache()
 
-    case OptionParser.parse(args, switches: @switches) do
-      {opts, [], _} ->
-        compile(filter_available_and_local_deps(deps), opts)
+    config = Mix.Project.config()
 
-      {opts, tail, _} ->
-        compile(Mix.Dep.filter_by_name(tail, deps, opts), opts)
-    end
+    Mix.Project.with_build_lock(config, fn ->
+      deps = Mix.Dep.load_and_cache()
+
+      case OptionParser.parse(args, switches: @switches) do
+        {opts, [], _} ->
+          compile(filter_available_and_local_deps(deps), opts)
+
+        {opts, tail, _} ->
+          compile(Mix.Dep.filter_by_name(tail, deps, opts), opts)
+      end
+    end)
   end
 
   @doc false
@@ -101,6 +106,23 @@ defmodule Mix.Tasks.Deps.Compile do
               false
           end
 
+        if compiled? do
+          build_path = Mix.Project.build_path(config)
+
+          lazy_message = fn ->
+            info = %{
+              app: dep.app,
+              scm: dep.scm,
+              manager: dep.manager,
+              os_pid: System.pid()
+            }
+
+            {:dep_compiled, info}
+          end
+
+          Mix.Sync.PubSub.broadcast(build_path, lazy_message)
+        end
+
         # We should touch fetchable dependencies even if they
         # did not compile otherwise they will always be marked
         # as stale, even when there is nothing to do.
@@ -180,6 +202,10 @@ defmodule Mix.Tasks.Deps.Compile do
   end
 
   defp do_rebar3(%Mix.Dep{opts: opts} = dep, config) do
+    if not Mix.Rebar.available?(:rebar3) do
+      handle_rebar_not_found(dep)
+    end
+
     dep_path = opts[:dest]
     build_path = opts[:build]
     File.mkdir_p!(build_path)
@@ -207,8 +233,14 @@ defmodule Mix.Tasks.Deps.Compile do
       {"TERM", "dumb"}
     ]
 
-    cmd = "#{rebar_cmd(dep)} bare compile --paths #{escape_path(lib_path)}"
-    do_command(dep, config, cmd, false, env)
+    {exec, args} = Mix.Rebar.rebar_args(:rebar3, ["bare", "compile", "--paths", lib_path])
+
+    if Mix.shell().cmd({exec, args}, opts_for_cmd(dep, config, env)) != 0 do
+      Mix.raise(
+        "Could not compile dependency #{inspect(dep.app)}, \"#{Enum.join([exec | args], " ")}\" command failed. " <>
+          deps_compile_feedback(dep.app)
+      )
+    end
 
     # Check if we have any new symlinks after compilation
     for dir <- ~w(include priv src),
@@ -220,21 +252,12 @@ defmodule Mix.Tasks.Deps.Compile do
     true
   end
 
-  defp escape_path(path) do
-    escape = if match?({:win32, _}, :os.type()), do: "^ ", else: "\\ "
-    String.replace(path, " ", escape)
-  end
-
   defp rebar_config(dep) do
     dep.extra
     |> Mix.Rebar.dependency_config()
     |> Mix.Rebar.serialize_config()
   end
 
-  defp rebar_cmd(%Mix.Dep{manager: manager} = dep) do
-    Mix.Rebar.rebar_cmd(manager) || handle_rebar_not_found(dep)
-  end
-
   defp handle_rebar_not_found(%Mix.Dep{app: app, manager: manager}) do
     shell = Mix.shell()
 
@@ -246,7 +269,7 @@ defmodule Mix.Tasks.Deps.Compile do
       "Shall I install #{manager}? (if running non-interactively, " <>
         "use \"mix local.rebar --force\")"
 
-    unless shell.yes?(install_question) do
+    if not shell.yes?(install_question) do
       error_message =
         "Could not find \"#{manager}\" to compile " <>
           "dependency #{inspect(app)}, please ensure \"#{manager}\" is available"
@@ -255,12 +278,15 @@ defmodule Mix.Tasks.Deps.Compile do
     end
 
     Mix.Tasks.Local.Rebar.run(["--force"])
-    Mix.Rebar.local_rebar_cmd(manager) || Mix.raise("\"#{manager}\" installation failed")
+
+    if not Mix.Rebar.available?(manager) do
+      Mix.raise("\"#{manager}\" installation failed")
+    end
   end
 
   defp do_make(dep, config) do
     command = make_command(dep)
-    do_command(dep, config, command, true, [{"IS_DEP", "1"}])
+    shell_cmd!(dep, config, command, [{"IS_DEP", "1"}])
     build_structure(dep, config)
     true
   end
@@ -289,7 +315,7 @@ defmodule Mix.Tasks.Deps.Compile do
 
   defp do_compile(%Mix.Dep{opts: opts} = dep, config) do
     if command = opts[:compile] do
-      do_command(dep, config, command, true)
+      shell_cmd!(dep, config, command)
       build_structure(dep, config)
       true
     else
@@ -297,19 +323,21 @@ defmodule Mix.Tasks.Deps.Compile do
     end
   end
 
-  defp do_command(dep, config, command, print_app?, env \\ []) do
-    %Mix.Dep{app: app, system_env: system_env, opts: opts} = dep
-
-    env = [{"ERL_LIBS", Path.join(config[:deps_build_path], "lib")} | system_env] ++ env
-
-    if Mix.shell().cmd(command, env: env, print_app: print_app?, cd: opts[:dest]) != 0 do
+  defp shell_cmd!(%Mix.Dep{app: app} = dep, config, command, env \\ []) do
+    if Mix.shell().cmd(command, [print_app: true] ++ opts_for_cmd(dep, config, env)) != 0 do
       Mix.raise(
         "Could not compile dependency #{inspect(app)}, \"#{command}\" command failed. " <>
           deps_compile_feedback(app)
       )
     end
 
-    true
+    :ok
+  end
+
+  defp opts_for_cmd(dep, config, env) do
+    %Mix.Dep{system_env: system_env, opts: opts} = dep
+    env = [{"ERL_LIBS", Path.join(config[:deps_build_path], "lib")} | system_env] ++ env
+    [env: env, cd: opts[:dest]]
   end
 
   defp build_structure(%Mix.Dep{opts: opts}, config) do
