@@ -512,7 +512,7 @@ defmodule Mix.Utils do
           do_symlink_or_copy(source, target, link)
 
         {:error, _} ->
-          unless File.dir?(target) do
+          if not File.dir?(target) do
             File.rm_rf!(target)
           end
 
@@ -692,6 +692,11 @@ defmodule Mix.Utils do
           :httpc.set_options([ipfamily: fallback(inet)], :mix)
           request |> httpc_request(http_options) |> httpc_response()
 
+        {:error, {:failed_connect, [{:to_address, _}, {inet, _, reason}]}}
+        when inet in [:inet, :inet6] and elem(reason, 0) == :tls_alert ->
+          http_options = put_in(http_options, [:ssl, :middlebox_comp_mode], false)
+          request |> httpc_request(http_options) |> httpc_response()
+
         response ->
           httpc_response(response)
       end
